import os
import pytest
from sqlobject import SQLObject, StringCol, IntCol
from sqlobject.sqlbuilder import Select, SOME
from sqlobject.tests.dbtest import getConnection, setupClass


########################################
# Test PosgreSQL sslmode
########################################


try:
    connection = getConnection()
except (AttributeError, NameError):
    # The module was imported during documentation building
    pass
else:
    if connection.dbName != "postgres":
        pytestmark = pytest.mark.skip("These tests require PostgreSQL")


class SOTestSSLMode(SQLObject):
    test = StringCol()


def test_sslmode():
    setupClass(SOTestSSLMode)
    connection = SOTestSSLMode._connection
    if (not connection.module.__name__.startswith('psycopg')) or \
            (os.name == 'nt'):
        pytest.skip("The test requires PostgreSQL, psycopg and ssl mode; "
                    "also it doesn't work on w32")

    connection = getConnection(sslmode='require')
    SOTestSSLMode._connection = connection
    test = SOTestSSLMode(test='test')  # Connect to the DB to test sslmode

    connection.cache.clear()
    test = SOTestSSLMode.select()[0]
    assert test.test == 'test'


########################################
# Test PosgreSQL list{Database,Tables}
########################################


class SOTestSOList(SQLObject):
    pass


def test_list_databases():
    assert connection.db in connection.listDatabases()


def test_list_tables():
    setupClass(SOTestSOList)
    assert SOTestSOList.sqlmeta.table in connection.listTables()


class SOTestSOME(SQLObject):
    value = IntCol()


def test_SOME():
    setupClass(SOTestSOME)
    SOTestSOME(value=10)
    SOTestSOME(value=20)
    SOTestSOME(value=30)
    assert len(list(SOTestSOME.select(
        SOTestSOME.q.value > SOME(Select([SOTestSOME.q.value]))))) == 2



from sqlobject import AND, ForeignKey, SQLMultipleJoin, SQLObject, \
    SQLRelatedJoin, StringCol
from sqlobject.tests.dbtest import inserts, setupClass

# Testing for expressing join, foreign keys,
# and instance identity in SQLBuilder expressions.


class SBPerson(SQLObject):
    name = StringCol()
    addresses = SQLMultipleJoin('SBAddress', joinColumn='personID')
    sharedAddresses = SQLRelatedJoin('SBAddress',
                                     addRemoveName='SharedAddress')


class SBAddress(SQLObject):
    city = StringCol()
    person = ForeignKey('SBPerson')
    sharedPeople = SQLRelatedJoin('SBPerson')


def setup_module(mod):
    global ppl
    setupClass([SBPerson, SBAddress])
    ppl = inserts(SBPerson, [('James',), ('Julia',)], 'name')
    adds = inserts(SBAddress, [('London', ppl[0].id),
                               ('Chicago', ppl[1].id),
                               ('Abu Dhabi', ppl[1].id)],
                   'city personID')
    ppl[0].addSharedAddress(adds[0])
    ppl[0].addSharedAddress(adds[1])
    ppl[1].addSharedAddress(adds[0])


def testJoin():
    assert list(SBPerson.select(
        AND(SBPerson.q.id == SBAddress.q.personID,
            SBAddress.q.city == 'London'))) == \
        list(SBAddress.selectBy(city='London').throughTo.person)

    assert list(SBAddress.select(
        AND(SBPerson.q.id == SBAddress.q.personID,
            SBPerson.q.name == 'Julia')).orderBy(SBAddress.q.city)) == \
        list(SBPerson.selectBy(name='Julia').
             throughTo.addresses.orderBy(SBAddress.q.city))


def testRelatedJoin():
    assert list(SBPerson.selectBy(name='Julia').throughTo.sharedAddresses) == \
        list(ppl[1].sharedAddresses)


def testInstance():
    assert list(SBAddress.select(
        AND(SBPerson.q.id == SBAddress.q.personID, SBPerson.q.id == ppl[0].id)
    )) == list(ppl[0].addresses)


def testFK():
    assert list(SBPerson.select(
        AND(SBAddress.j.person, SBAddress.q.city == 'London'))) == \
        list(SBPerson.select(
            AND(SBPerson.q.id == SBAddress.q.personID,
                SBAddress.q.city == 'London')))


def testRelatedJoin2():
    assert list(SBAddress.select(
        AND(SBAddress.j.sharedPeople, SBPerson.q.name == 'Julia'))) == \
        list(SBPerson.select(
             SBPerson.q.name == 'Julia').throughTo.sharedAddresses)


def testJoin2():
    assert list(SBAddress.select(
        AND(SBPerson.j.addresses,
            SBPerson.q.name == 'Julia')).orderBy(SBAddress.q.city)) == \
        list(SBAddress.select(
            AND(SBPerson.q.id == SBAddress.q.personID,
                SBPerson.q.name == 'Julia')).orderBy(SBAddress.q.city)) == \
        list(SBPerson.selectBy(name='Julia').throughTo.
             addresses.orderBy(SBAddress.q.city))


def testFK2():
    assert list(SBAddress.select(
        AND(SBAddress.j.person, SBPerson.q.name == 'Julia'))) == \
        list(SBAddress.select(
            AND(SBPerson.q.id == SBAddress.q.personID,
                SBPerson.q.name == 'Julia')))



from sqlobject import IntCol, SQLObject, StringCol
from sqlobject.sqlbuilder import Select, func
from sqlobject.tests.dbtest import getConnection, setupClass


########################################
# groupBy
########################################


class GroupbyTest(SQLObject):
    name = StringCol()
    so_value = IntCol()


def test_groupBy():
    setupClass(GroupbyTest)
    GroupbyTest(name='a', so_value=1)
    GroupbyTest(name='a', so_value=2)
    GroupbyTest(name='b', so_value=1)

    connection = getConnection()
    select = Select(
        [GroupbyTest.q.name, func.COUNT(GroupbyTest.q.so_value)],
        groupBy=GroupbyTest.q.name,
        orderBy=GroupbyTest.q.name)
    sql = connection.sqlrepr(select)
    rows = list(connection.queryAll(sql))
    assert [tuple(t) for t in rows] == [('a', 2), ('b', 1)]


def test_groupBy_list():
    setupClass(GroupbyTest)
    GroupbyTest(name='a', so_value=1)
    GroupbyTest(name='a', so_value=2)
    GroupbyTest(name='b', so_value=1)

    connection = getConnection()
    select = Select(
        [GroupbyTest.q.name, GroupbyTest.q.so_value],
        groupBy=[GroupbyTest.q.name, GroupbyTest.q.so_value],
        orderBy=[GroupbyTest.q.name, GroupbyTest.q.so_value])
    sql = connection.sqlrepr(select)
    rows = list(connection.queryAll(sql))
    assert [tuple(t) for t in rows] == [('a', 1), ('a', 2), ('b', 1)]



from sqlobject import SQLObject, JSONCol
from sqlobject.tests.dbtest import setupClass


class JSONTest(SQLObject):
    json = JSONCol(default=None)


_json_test_data = (
    None, True, 1, 2.0,
    {u"test": [None, True, 1, 2.0,
     {u"unicode'with'apostrophes": u"unicode\"with\"quotes"},
     [],
     u"unicode"]},
    [None, True, 1, 2.0,
     [],
     {u"unicode'with'apostrophes": u"unicode\"with\"quotes"},
     u"unicode", u"unicode'with'apostrophes", u"unicode\"with\"quotes",
     ],
    u"unicode", u"unicode'with'apostrophes", u"unicode\"with\"quotes",
)


def test_JSONCol():
    setupClass(JSONTest)

    for _id, test_data in enumerate(_json_test_data):
        json = JSONTest(id=_id + 1, json=test_data)

    JSONTest._connection.cache.clear()

    for _id, test_data in enumerate(_json_test_data):
        json = JSONTest.get(_id + 1)
        assert json.json == test_data



import os
import sys
import types

# Credit to six authors: https://pypi.org/project/six/
# License: MIT


def with_metaclass(meta, *bases):
    """Create a base class with a metaclass."""
    # This requires a bit of explanation: the basic idea is to make a dummy
    # metaclass for one level of class instantiation that replaces itself with
    # the actual metaclass.

    class metaclass(meta):
        def __new__(cls, name, this_bases, d):
            return meta(name, bases, d)
    return type.__new__(metaclass, 'temporary_class', (), {})

# Compatability definitions (inspired by six)
PY2 = sys.version_info[0] < 3
if PY2:
    # disable flake8 checks on python 3
    string_type = basestring  # noqa
    unicode_type = unicode  # noqa
    class_types = (type, types.ClassType)
    buffer_type = buffer  # noqa
else:
    string_type = str
    unicode_type = str
    class_types = (type,)
    buffer_type = memoryview

if PY2:
    import imp

    def load_module_from_file(base_name, module_name, filename):
        fp, pathname, description = imp.find_module(
            base_name, [os.path.dirname(filename)])
        try:
            module = imp.load_module(module_name, fp, pathname, description)
        finally:
            fp.close()
        return module
else:
    import importlib.util

    def load_module_from_file(base_name, module_name, filename):
        specs = importlib.util.spec_from_file_location(module_name, filename)
        return specs.loader.load_module()



from sqlobject import ForeignKey, MultipleJoin, RelatedJoin, SQLObject, \
    StringCol
from sqlobject.tests.dbtest import setupClass


########################################
# Joins
########################################


class PersonJoiner(SQLObject):

    name = StringCol(length=40, alternateID=True)
    addressJoiners = RelatedJoin('AddressJoiner')


class AddressJoiner(SQLObject):

    zip = StringCol(length=5, alternateID=True)
    personJoiners = RelatedJoin('PersonJoiner')


class ImplicitJoiningSO(SQLObject):
    foo = RelatedJoin('Bar')


class ExplicitJoiningSO(SQLObject):
    foo = MultipleJoin('Bar', joinMethodName='foo')


class TestJoin:

    def setup_method(self, meth):
        setupClass(PersonJoiner)
        setupClass(AddressJoiner)
        for n in ['bob', 'tim', 'jane', 'joe', 'fred', 'barb']:
            PersonJoiner(name=n)
        for z in ['11111', '22222', '33333', '44444']:
            AddressJoiner(zip=z)

    def test_join(self):
        b = PersonJoiner.byName('bob')
        assert b.addressJoiners == []
        z = AddressJoiner.byZip('11111')
        b.addAddressJoiner(z)
        self.assertZipsEqual(b.addressJoiners, ['11111'])
        self.assertNamesEqual(z.personJoiners, ['bob'])
        z2 = AddressJoiner.byZip('22222')
        b.addAddressJoiner(z2)
        self.assertZipsEqual(b.addressJoiners, ['11111', '22222'])
        self.assertNamesEqual(z2.personJoiners, ['bob'])
        b.removeAddressJoiner(z)
        self.assertZipsEqual(b.addressJoiners, ['22222'])
        self.assertNamesEqual(z.personJoiners, [])

    def assertZipsEqual(self, zips, dest):
        assert [a.zip for a in zips] == dest

    def assertNamesEqual(self, people, dest):
        assert [p.name for p in people] == dest

    def test_joinAttributeWithUnderscores(self):
        # Make sure that the implicit setting of joinMethodName works
        assert hasattr(ImplicitJoiningSO, 'foo')
        assert not hasattr(ImplicitJoiningSO, 'bars')

        # And make sure explicit setting also works
        assert hasattr(ExplicitJoiningSO, 'foo')
        assert not hasattr(ExplicitJoiningSO, 'bars')


class PersonJoiner2(SQLObject):

    name = StringCol('name', length=40, alternateID=True)
    addressJoiner2s = MultipleJoin('AddressJoiner2')


class AddressJoiner2(SQLObject):

    class sqlmeta:
        defaultOrder = ['-zip', 'plus4']

    zip = StringCol(length=5)
    plus4 = StringCol(length=4, default=None)
    personJoiner2 = ForeignKey('PersonJoiner2')


class TestJoin2:

    def setup_method(self, meth):
        setupClass([PersonJoiner2, AddressJoiner2])
        p1 = PersonJoiner2(name='bob')
        p2 = PersonJoiner2(name='sally')
        for z in ['11111', '22222', '33333']:
            AddressJoiner2(zip=z, personJoiner2=p1)
        AddressJoiner2(zip='00000', personJoiner2=p2)

    def test_basic(self):
        bob = PersonJoiner2.byName('bob')
        sally = PersonJoiner2.byName('sally')
        assert len(bob.addressJoiner2s) == 3
        assert len(sally.addressJoiner2s) == 1
        bob.addressJoiner2s[0].destroySelf()
        assert len(bob.addressJoiner2s) == 2
        z = bob.addressJoiner2s[0]
        z.zip = 'xxxxx'
        id = z.id
        del z
        z = AddressJoiner2.get(id)
        assert z.zip == 'xxxxx'

    def test_defaultOrder(self):
        p1 = PersonJoiner2.byName('bob')
        assert ([i.zip for i in p1.addressJoiner2s]
                == ['33333', '22222', '11111'])


_personJoiner3_getters = []
_personJoiner3_setters = []


class PersonJoiner3(SQLObject):

    name = StringCol('name', length=40, alternateID=True)
    addressJoiner3s = MultipleJoin('AddressJoiner3')


class AddressJoiner3(SQLObject):

    zip = StringCol(length=5)
    personJoiner3 = ForeignKey('PersonJoiner3')

    def _get_personJoiner3(self):
        value = self._SO_get_personJoiner3()
        _personJoiner3_getters.append((self, value))
        return value

    def _set_personJoiner3(self, value):
        self._SO_set_personJoiner3(value)
        _personJoiner3_setters.append((self, value))


class TestJoin3:

    def setup_method(self, meth):
        setupClass([PersonJoiner3, AddressJoiner3])
        p1 = PersonJoiner3(name='bob')
        p2 = PersonJoiner3(name='sally')
        for z in ['11111', '22222', '33333']:
            AddressJoiner3(zip=z, personJoiner3=p1)
        AddressJoiner3(zip='00000', personJoiner3=p2)

    def test_accessors(self):
        assert len(_personJoiner3_getters) == 0
        assert len(_personJoiner3_setters) == 4
        bob = PersonJoiner3.byName('bob')
        for addressJoiner3 in bob.addressJoiner3s:
            addressJoiner3.personJoiner3
        assert len(_personJoiner3_getters) == 3
        assert len(_personJoiner3_setters) == 4



from sqlobject.declarative import Declarative


class A1(Declarative):

    a = 1
    b = []


class A2(A1):

    a = 5

A3 = A2(b=5)


def test_a_classes():
    assert A1.a == 1
    assert A1.singleton().a == 1
    assert A1.b is A2.b
    assert A3.b == 5
    assert A1.declarative_count == A1.singleton().declarative_count
    assert A1.declarative_count < A2.declarative_count
    assert A2.singleton() is not A1.singleton()
    assert A3.singleton().b == A3.b


class B1(Declarative):

    attrs = []

    def __classinit__(cls, new_attrs):
        Declarative.__classinit__(cls, new_attrs)
        cls.attrs = cls.add_attrs(cls.attrs, new_attrs)

    def __instanceinit__(self, new_attrs):
        Declarative.__instanceinit__(self, new_attrs)
        self.attrs = self.add_attrs(self.attrs, new_attrs)

    @staticmethod
    def add_attrs(old_attrs, new_attrs):
        old_attrs = old_attrs[:]
        for name in new_attrs.keys():
            if (name in old_attrs or name.startswith('_')
                    or name in ('add_attrs', 'declarative_count', 'attrs')):
                continue
            old_attrs.append(name)
        old_attrs.sort()
        return old_attrs

    c = 1


class B2(B1):

    g = 3

    def __classinit__(cls, new_attrs):
        new_attrs['test'] = 'whatever'
        B1.__classinit__(cls, new_attrs)

B3 = B2(c=5, d=3)
B4 = B3(d=5)
B5 = B1(a=1)


def test_b_classes():
    assert B1.attrs == ['c']
    assert B1.c == 1
    assert B2.attrs == ['c', 'g', 'test']
    assert B3.d == 3
    assert B4.d == 5
    assert B5.a == 1
    assert B5.attrs == ['a', 'c']
    assert B3.attrs == ['c', 'd', 'g', 'test']
    assert B4.attrs == ['c', 'd', 'g', 'test']
    order = [B1, B1.singleton(), B2, B2.singleton(),
             B3, B3.singleton(), B4, B4.singleton(),
             B5, B5.singleton()]
    last = 0
    for obj in order:
        assert obj.declarative_count >= last
        last = obj.declarative_count



"""
SQLObject
---------

:author: Ian Bicking <ianb@colorstudy.com>

SQLObject is a object-relational mapper.  See SQLObject.html or
SQLObject.rst for more.

With the help by Oleg Broytman and many other contributors.
See Authors.rst.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as
published by the Free Software Foundation; either version 2.1 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301,
USA.
"""
import sys
import threading
import weakref
import types
import warnings
from . import sqlbuilder
from . import dbconnection
from . import col
from . import styles
from . import joins
from . import index
from . import classregistry
from . import declarative
from . import events
from .sresults import SelectResults
from .util.threadinglocal import local
from sqlobject.compat import PY2, with_metaclass, string_type, unicode_type

if ((sys.version_info[0] == 2) and (sys.version_info[:2] < (2, 7))) or \
   ((sys.version_info[0] == 3) and (sys.version_info[:2] < (3, 4))):
    raise ImportError("SQLObject requires Python 2.7 or 3.4+")

if not PY2:
    # alias for python 3 compatability
    long = int

"""
This thread-local storage is needed for RowCreatedSignals. It gathers
code-blocks to execute _after_ the whole hierachy of inherited SQLObjects
is created. See SQLObject._create
"""

NoDefault = sqlbuilder.NoDefault


class SQLObjectNotFound(LookupError):
    pass


class SQLObjectIntegrityError(Exception):
    pass


def makeProperties(obj):
    """
    This function takes a dictionary of methods and finds
    methods named like:
    * _get_attr
    * _set_attr
    * _del_attr
    * _doc_attr
    Except for _doc_attr, these should be methods.  It
    then creates properties from these methods, like
    property(_get_attr, _set_attr, _del_attr, _doc_attr).
    Missing methods are okay.
    """

    if isinstance(obj, dict):
        def setFunc(var, value):
            obj[var] = value
        d = obj
    else:
        def setFunc(var, value):
            setattr(obj, var, value)
        d = obj.__dict__

    props = {}
    for var, value in d.items():
        if var.startswith('_set_'):
            props.setdefault(var[5:], {})['set'] = value
        elif var.startswith('_get_'):
            props.setdefault(var[5:], {})['get'] = value
        elif var.startswith('_del_'):
            props.setdefault(var[5:], {})['del'] = value
        elif var.startswith('_doc_'):
            props.setdefault(var[5:], {})['doc'] = value
    for var, setters in props.items():
        if len(setters) == 1 and 'doc' in setters:
            continue
        if var in d:
            if isinstance(d[var], (types.MethodType, types.FunctionType)):
                warnings.warn(
                    "I tried to set the property %r, but it was "
                    "already set, as a method (%r).  Methods have "
                    "significantly different semantics than properties, "
                    "and this may be a sign of a bug in your code."
                    % (var, d[var]))
            continue
        setFunc(var,
                property(setters.get('get'), setters.get('set'),
                         setters.get('del'), setters.get('doc')))


def unmakeProperties(obj):
    if isinstance(obj, dict):
        def delFunc(obj, var):
            del obj[var]
        d = obj
    else:
        delFunc = delattr
        d = obj.__dict__

    for var, value in list(d.items()):
        if isinstance(value, property):
            for prop in [value.fget, value.fset, value.fdel]:
                if prop and prop.__name__ not in d:
                    delFunc(obj, var)
                    break


def findDependencies(name, registry=None):
    depends = []
    for klass in classregistry.registry(registry).allClasses():
        if findDependantColumns(name, klass):
            depends.append(klass)
        else:
            for join in klass.sqlmeta.joins:
                if isinstance(join, joins.SORelatedJoin) and \
                        join.otherClassName == name:
                    depends.append(klass)
                    break
    return depends


def findDependantColumns(name, klass):
    depends = []
    for _col in klass.sqlmeta.columnList:
        if _col.foreignKey == name and _col.cascade is not None:
            depends.append(_col)
    return depends


def _collectAttributes(cls, new_attrs, look_for_class):
    """Finds all attributes in `new_attrs` that are instances of
    `look_for_class`. The ``.name`` attribute is set for any matching objects.
    Returns them as a list.

    """
    result = []
    for attr, value in new_attrs.items():
        if isinstance(value, look_for_class):
            value.name = attr
            delattr(cls, attr)
            result.append(value)
    return result


class CreateNewSQLObject:
    """
    Dummy singleton to use in place of an ID, to signal we want
    a new object.
    """
    pass


class sqlmeta(with_metaclass(declarative.DeclarativeMeta, object)):
    """
    This object is the object we use to keep track of all sorts of
    information.  Subclasses are made for each SQLObject subclass
    (dynamically if necessary), and instances are created to go
    alongside every SQLObject instance.
    """

    table = None
    idName = None
    idSequence = None
    # This function is used to coerce IDs into the proper format,
    # so you should replace it with str, or another function, if you
    # aren't using integer IDs
    idType = int
    style = None
    lazyUpdate = False
    defaultOrder = None
    cacheValues = True
    registry = None
    fromDatabase = False
    # Default is false, but we set it to true for the *instance*
    # when necessary: (bad clever? maybe)
    expired = False

    # This is a mapping from column names to SOCol (or subclass)
    # instances:
    columns = {}
    columnList = []

    # This is a mapping from column names to Col (or subclass)
    # instances; these objects don't have the logic that the SOCol
    # objects do, and are not attached to this class closely.
    columnDefinitions = {}

    # These are lists of the join and index objects:
    indexes = []
    indexDefinitions = []
    joins = []
    joinDefinitions = []

    # These attributes shouldn't be shared with superclasses:
    _unshared_attributes = ['table', 'columns', 'childName']

    # These are internal bookkeeping attributes; the class-level
    # definition is a default for the instances, instances will
    # reset these values.

    # When an object is being created, it has an instance
    # variable _creating, which is true.  This way all the
    # setters can be captured until the object is complete,
    # and then the row is inserted into the database.  Once
    # that happens, _creating is deleted from the instance,
    # and only the class variable (which is always false) is
    # left.
    _creating = False
    _obsolete = False
    # Sometimes an intance is attached to a connection, not
    # globally available.  In that case, self.sqlmeta._perConnection
    # will be true.  It's false by default:
    _perConnection = False

    # Inheritance definitions:
    parentClass = None  # A reference to the parent class
    childClasses = {}  # References to child classes, keyed by childName
    childName = None  # Class name for inheritance child object creation

    # Does the row require syncing?
    dirty = False

    # Default encoding for UnicodeCol's
    dbEncoding = None

    def __classinit__(cls, new_attrs):
        for attr in cls._unshared_attributes:
            if attr not in new_attrs:
                setattr(cls, attr, None)
        declarative.setup_attributes(cls, new_attrs)

    def __init__(self, instance):
        self.instance = weakref.proxy(instance)

    @classmethod
    def send(cls, signal, *args, **kw):
        events.send(signal, cls.soClass, *args, **kw)

    @classmethod
    def setClass(cls, soClass):
        cls.soClass = soClass
        if not cls.style:
            cls.style = styles.defaultStyle
            try:
                if cls.soClass._connection and cls.soClass._connection.style:
                    cls.style = cls.soClass._connection.style
            except AttributeError:
                pass
        if cls.table is None:
            cls.table = cls.style.pythonClassToDBTable(cls.soClass.__name__)
        if cls.idName is None:
            cls.idName = cls.style.idForTable(cls.table)

        # plainSetters are columns that haven't been overridden by the
        # user, so we can contact the database directly to set them.
        # Note that these can't set these in the SQLObject class
        # itself, because they specific to this subclass of SQLObject,
        # and cannot be shared among classes.
        cls._plainSetters = {}
        cls._plainGetters = {}
        cls._plainForeignSetters = {}
        cls._plainForeignGetters = {}
        cls._plainJoinGetters = {}
        cls._plainJoinAdders = {}
        cls._plainJoinRemovers = {}

        # This is a dictionary of columnName: columnObject
        # None of these objects can be shared with superclasses
        cls.columns = {}
        cls.columnList = []
        # These, however, can be shared:
        cls.columnDefinitions = cls.columnDefinitions.copy()
        cls.indexes = []
        cls.indexDefinitions = cls.indexDefinitions[:]
        cls.joins = []
        cls.joinDefinitions = cls.joinDefinitions[:]

    ############################################################
    # Adding special values, like columns and indexes
    ############################################################

    ########################################
    # Column handling
    ########################################

    @classmethod
    def addColumn(cls, columnDef, changeSchema=False, connection=None):
        post_funcs = []
        cls.send(events.AddColumnSignal, cls.soClass, connection,
                 columnDef.name, columnDef, changeSchema, post_funcs)
        sqlmeta = cls
        soClass = cls.soClass
        del cls
        column = columnDef.withClass(soClass)
        name = column.name
        assert name != 'id', (
            "The 'id' column is implicit, and should not be defined as "
            "a column")
        assert name not in sqlmeta.columns, (
            "The class %s.%s already has a column %r (%r), you cannot "
            "add the column %r"
            % (soClass.__module__, soClass.__name__, name,
               sqlmeta.columnDefinitions[name], columnDef))
        # Collect columns from the parent classes to test
        # if the column is not in a parent class
        parent_columns = []
        for base in soClass.__bases__:
            if hasattr(base, "sqlmeta"):
                parent_columns.extend(base.sqlmeta.columns.keys())
        if hasattr(soClass, name):
            assert (name in parent_columns) or (name == "childName"), (
                "The class %s.%s already has a variable or method %r, "
                "you cannot add the column %r" % (
                    soClass.__module__, soClass.__name__, name, name))
        sqlmeta.columnDefinitions[name] = columnDef
        sqlmeta.columns[name] = column
        # A stable-ordered version of the list...
        sqlmeta.columnList.append(column)

        ###################################################
        # Create the getter function(s).  We'll start by
        # creating functions like _SO_get_columnName,
        # then if there's no function named _get_columnName
        # we'll alias that to _SO_get_columnName.  This
        # allows a sort of super call, even though there's
        # no superclass that defines the database access.
        if sqlmeta.cacheValues:
            # We create a method here, which is just a function
            # that takes "self" as the first argument.
            getter = eval(
                'lambda self: self._SO_loadValue(%s)' %
                repr(instanceName(name)))

        else:
            # If we aren't caching values, we just call the
            # function _SO_getValue, which fetches from the
            # database.
            getter = eval('lambda self: self._SO_getValue(%s)' % repr(name))
        setattr(soClass, rawGetterName(name), getter)

        # Here if the _get_columnName method isn't in the
        # definition, we add it with the default
        # _SO_get_columnName definition.
        if not hasattr(soClass, getterName(name)) or (name == 'childName'):
            setattr(soClass, getterName(name), getter)
            sqlmeta._plainGetters[name] = 1

        #################################################
        # Create the setter function(s)
        # Much like creating the getters, we will create
        # _SO_set_columnName methods, and then alias them
        # to _set_columnName if the user hasn't defined
        # those methods themself.

        # @@: This is lame; immutable right now makes it unsettable,
        # making the table read-only
        if not column.immutable:
            # We start by just using the _SO_setValue method
            setter = eval(
                'lambda self, val: self._SO_setValue'
                '(%s, val, self.%s, self.%s)' % (
                    repr(name),
                    '_SO_from_python_%s' % name, '_SO_to_python_%s' % name))
            setattr(soClass, '_SO_from_python_%s' % name, column.from_python)
            setattr(soClass, '_SO_to_python_%s' % name, column.to_python)
            setattr(soClass, rawSetterName(name), setter)
            # Then do the aliasing
            if not hasattr(soClass, setterName(name)) or (name == 'childName'):
                setattr(soClass, setterName(name), setter)
                # We keep track of setters that haven't been
                # overridden, because we can combine these
                # set columns into one SQL UPDATE query.
                sqlmeta._plainSetters[name] = 1

        ##################################################
        # Here we check if the column is a foreign key, in
        # which case we need to make another method that
        # fetches the key and constructs the sister
        # SQLObject instance.
        if column.foreignKey:

            # We go through the standard _SO_get_columnName deal
            # we're giving the object, not the ID of the
            # object this time:
            origName = column.origName
            if sqlmeta.cacheValues:
                # self._SO_class_className is a reference
                # to the class in question.
                getter = eval(
                    'lambda self: self._SO_foreignKey'
                    '(self._SO_loadValue(%r), self._SO_class_%s, %s)' %
                    (instanceName(name), column.foreignKey,
                     column.refColumn and repr(column.refColumn)))
            else:
                # Same non-caching version as above.
                getter = eval(
                    'lambda self: self._SO_foreignKey'
                    '(self._SO_getValue(%s), self._SO_class_%s, %s)' %
                    (repr(name), column.foreignKey,
                     column.refColumn and repr(column.refColumn)))
            setattr(soClass, rawGetterName(origName), getter)

            # And we set the _get_columnName version
            if not hasattr(soClass, getterName(origName)):
                setattr(soClass, getterName(origName), getter)
                sqlmeta._plainForeignGetters[origName] = 1

            if not column.immutable:
                # The setter just gets the ID of the object,
                # and then sets the real column.
                setter = eval(
                    'lambda self, val: '
                    'setattr(self, %s, self._SO_getID(val, %s))' %
                    (repr(name), column.refColumn and repr(column.refColumn)))
                setattr(soClass, rawSetterName(origName), setter)
                if not hasattr(soClass, setterName(origName)):
                    setattr(soClass, setterName(origName), setter)
                    sqlmeta._plainForeignSetters[origName] = 1

            classregistry.registry(sqlmeta.registry).addClassCallback(
                column.foreignKey,
                lambda foreign, me, attr: setattr(me, attr, foreign),
                soClass, '_SO_class_%s' % column.foreignKey)

        if column.alternateMethodName:
            func = eval(
                'lambda cls, val, connection=None: '
                'cls._SO_fetchAlternateID(%s, %s, val, connection=connection)'
                % (repr(column.name), repr(column.dbName)))
            setattr(soClass, column.alternateMethodName, classmethod(func))

        if changeSchema:
            conn = connection or soClass._connection
            conn.addColumn(sqlmeta.table, column)

        if soClass._SO_finishedClassCreation:
            makeProperties(soClass)

        for func in post_funcs:
            func(soClass, column)

    @classmethod
    def addColumnsFromDatabase(sqlmeta, connection=None):
        soClass = sqlmeta.soClass
        conn = connection or soClass._connection
        for columnDef in conn.columnsFromSchema(sqlmeta.table, soClass):
            if columnDef.name not in sqlmeta.columnDefinitions:
                if isinstance(columnDef.name, unicode_type) and PY2:
                    columnDef.name = columnDef.name.encode('ascii')
                sqlmeta.addColumn(columnDef)

    @classmethod
    def delColumn(cls, column, changeSchema=False, connection=None):
        sqlmeta = cls
        soClass = sqlmeta.soClass
        if isinstance(column, str):
            if column in sqlmeta.columns:
                column = sqlmeta.columns[column]
            elif column + 'ID' in sqlmeta.columns:
                column = sqlmeta.columns[column + 'ID']
            else:
                raise ValueError('Unknown column ' + column)
        if isinstance(column, col.Col):
            for c in sqlmeta.columns.values():
                if column is c.columnDef:
                    column = c
                    break
            else:
                raise IndexError(
                    "Column with definition %r not found" % column)
        post_funcs = []
        cls.send(events.DeleteColumnSignal, cls.soClass, connection,
                 column.name, column, post_funcs)
        name = column.name
        del sqlmeta.columns[name]
        del sqlmeta.columnDefinitions[name]
        sqlmeta.columnList.remove(column)
        delattr(soClass, rawGetterName(name))
        if name in sqlmeta._plainGetters:
            delattr(soClass, getterName(name))
        delattr(soClass, rawSetterName(name))
        if name in sqlmeta._plainSetters:
            delattr(soClass, setterName(name))
        if column.foreignKey:
            delattr(soClass,
                    rawGetterName(soClass.sqlmeta.style.
                                  instanceIDAttrToAttr(name)))
            if name in sqlmeta._plainForeignGetters:
                delattr(soClass, getterName(name))
            delattr(soClass,
                    rawSetterName(soClass.sqlmeta.style.
                                  instanceIDAttrToAttr(name)))
            if name in sqlmeta._plainForeignSetters:
                delattr(soClass, setterName(name))
        if column.alternateMethodName:
            delattr(soClass, column.alternateMethodName)

        if changeSchema:
            conn = connection or soClass._connection
            conn.delColumn(sqlmeta, column)

        if soClass._SO_finishedClassCreation:
            unmakeProperties(soClass)
            makeProperties(soClass)

        for func in post_funcs:
            func(soClass, column)

    ########################################
    # Join handling
    ########################################

    @classmethod
    def addJoin(cls, joinDef):
        sqlmeta = cls
        soClass = cls.soClass
        # The name of the method we'll create.  If it's
        # automatically generated, it's generated by the
        # join class.
        join = joinDef.withClass(soClass)
        meth = join.joinMethodName

        sqlmeta.joins.append(join)
        index = len(sqlmeta.joins) - 1
        if joinDef not in sqlmeta.joinDefinitions:
            sqlmeta.joinDefinitions.append(joinDef)

        # The function fetches the join by index, and
        # then lets the join object do the rest of the
        # work:
        func = eval(
            'lambda self: self.sqlmeta.joins[%i].performJoin(self)' % index)

        # And we do the standard _SO_get_... _get_... deal
        setattr(soClass, rawGetterName(meth), func)
        if not hasattr(soClass, getterName(meth)):
            setattr(soClass, getterName(meth), func)
            sqlmeta._plainJoinGetters[meth] = 1

        # Some joins allow you to remove objects from the
        # join.
        if hasattr(join, 'remove'):
            # Again, we let it do the remove, and we do the
            # standard naming trick.
            func = eval(
                'lambda self, obj: self.sqlmeta.joins[%i].remove(self, obj)' %
                index)
            setattr(soClass, '_SO_remove' + join.addRemoveName, func)
            if not hasattr(soClass, 'remove' + join.addRemoveName):
                setattr(soClass, 'remove' + join.addRemoveName, func)
                sqlmeta._plainJoinRemovers[meth] = 1

        # Some joins allow you to add objects.
        if hasattr(join, 'add'):
            # And again...
            func = eval(
                'lambda self, obj: self.sqlmeta.joins[%i].add(self, obj)' %
                index)
            setattr(soClass, '_SO_add' + join.addRemoveName, func)
            if not hasattr(soClass, 'add' + join.addRemoveName):
                setattr(soClass, 'add' + join.addRemoveName, func)
                sqlmeta._plainJoinAdders[meth] = 1

        if soClass._SO_finishedClassCreation:
            makeProperties(soClass)

    @classmethod
    def delJoin(sqlmeta, joinDef):
        soClass = sqlmeta.soClass
        for join in sqlmeta.joins:
            # previously deleted joins will be None, so it must
            # be skipped or it'll error out on the next line.
            if join is None:
                continue
            if joinDef is join.joinDef:
                break
        else:
            raise IndexError(
                "Join %r not found in class %r (from %r)"
                % (joinDef, soClass, sqlmeta.joins))
        meth = join.joinMethodName
        sqlmeta.joinDefinitions.remove(joinDef)
        for i in range(len(sqlmeta.joins)):
            if sqlmeta.joins[i] is join:
                # Have to leave None, because we refer to joins
                # by index.
                sqlmeta.joins[i] = None
        delattr(soClass, rawGetterName(meth))
        if meth in sqlmeta._plainJoinGetters:
            delattr(soClass, getterName(meth))
        if hasattr(join, 'remove'):
            delattr(soClass, '_SO_remove' + join.addRemovePrefix)
            if meth in sqlmeta._plainJoinRemovers:
                delattr(soClass, 'remove' + join.addRemovePrefix)
        if hasattr(join, 'add'):
            delattr(soClass, '_SO_add' + join.addRemovePrefix)
            if meth in sqlmeta._plainJoinAdders:
                delattr(soClass, 'add' + join.addRemovePrefix)

        if soClass._SO_finishedClassCreation:
            unmakeProperties(soClass)
            makeProperties(soClass)

    ########################################
    # Indexes
    ########################################

    @classmethod
    def addIndex(cls, indexDef):
        cls.indexDefinitions.append(indexDef)
        index = indexDef.withClass(cls.soClass)
        cls.indexes.append(index)
        setattr(cls.soClass, index.name, index)

    ########################################
    # Utility methods
    ########################################

    @classmethod
    def getColumns(sqlmeta):
        return sqlmeta.columns.copy()

    def asDict(self):
        """
        Return the object as a dictionary of columns to values.
        """
        result = {}
        for key in self.getColumns():
            result[key] = getattr(self.instance, key)
        result['id'] = self.instance.id
        return result

    @classmethod
    def expireAll(sqlmeta, connection=None):
        """
        Expire all instances of this class.
        """
        soClass = sqlmeta.soClass
        connection = connection or soClass._connection
        cache_set = connection.cache
        cache_set.weakrefAll(soClass)
        for item in cache_set.getAll(soClass):
            item.expire()


sqlhub = dbconnection.ConnectionHub()


# Turning it on gives earlier warning about things
# that will be deprecated (having this off we won't flood people
# with warnings right away).
warnings_level = 1
exception_level = None
# Current levels:
#  1) Actively deprecated
#  2) Deprecated after 1
#  3) Deprecated after 2


def deprecated(message, level=1, stacklevel=2):
    if exception_level is not None and exception_level <= level:
        raise NotImplementedError(message)
    if warnings_level is not None and warnings_level <= level:
        warnings.warn(message, DeprecationWarning, stacklevel=stacklevel)

# if sys.version_info[:2] < (2, 7):
#     deprecated("Support for Python 2.6 has been declared obsolete "
#                "and will be removed in the next release of SQLObject")


def setDeprecationLevel(warning=1, exception=None):
    """
    Set the deprecation level for SQLObject.  Low levels are more
    actively being deprecated.  Any warning at a level at or below
    ``warning`` will give a warning.  Any warning at a level at or
    below ``exception`` will give an exception.  You can use a higher
    ``exception`` level for tests to help upgrade your code.  ``None``
    for either value means never warn or raise exceptions.

    The levels currently mean:

      1) Deprecated in current version.  Will be removed in next version.

      2) Planned to deprecate in next version, remove later.

      3) Planned to deprecate sometime, remove sometime much later.

    As the SQLObject versions progress, the deprecation level of
    specific features will go down, indicating the advancing nature of
    the feature's doom.  We'll try to keep features at 1 for a major
    revision.

    As time continues there may be a level 0, which will give a useful
    error message (better than ``AttributeError``) but where the
    feature has been fully removed.
    """
    global warnings_level, exception_level
    warnings_level = warning
    exception_level = exception


class _sqlmeta_attr(object):

    def __init__(self, name, deprecation_level):
        self.name = name
        self.deprecation_level = deprecation_level

    def __get__(self, obj, type=None):
        if self.deprecation_level is not None:
            deprecated(
                'Use of this attribute should be replaced with '
                '.sqlmeta.%s' % self.name, level=self.deprecation_level)
        return getattr((type or obj).sqlmeta, self.name)


_postponed_local = local()


# SQLObject is the superclass for all SQLObject classes, of
# course.  All the deeper magic is done in MetaSQLObject, and
# only lesser magic is done here.  All the actual work is done
# here, though -- just automatic method generation (like
# methods and properties for each column) is done in
# MetaSQLObject.


class SQLObject(with_metaclass(declarative.DeclarativeMeta, object)):

    _connection = sqlhub

    sqlmeta = sqlmeta

    # DSM: The _inheritable attribute controls wheter the class can by
    # DSM: inherited 'logically' with a foreignKey and a back reference.
    _inheritable = False  # Is this class inheritable?
    _parent = None  # A reference to the parent instance
    childName = None  # Children name (to be able to get a subclass)

    # The law of Demeter: the class should not call another classes by name
    SelectResultsClass = SelectResults

    def __classinit__(cls, new_attrs):

        # This is true if we're initializing the SQLObject class,
        # instead of a subclass:
        is_base = cls.__bases__ == (object,)

        cls._SO_setupSqlmeta(new_attrs, is_base)

        implicitColumns = _collectAttributes(cls, new_attrs, col.Col)
        implicitJoins = _collectAttributes(cls, new_attrs, joins.Join)
        implicitIndexes = _collectAttributes(cls, new_attrs,
                                             index.DatabaseIndex)

        if not is_base:
            cls._SO_cleanDeprecatedAttrs(new_attrs)

        if '_connection' in new_attrs:
            connection = new_attrs['_connection']
            del cls._connection
            assert 'connection' not in new_attrs
        elif 'connection' in new_attrs:
            connection = new_attrs['connection']
            del cls.connection
        else:
            connection = None

        cls._SO_finishedClassCreation = False

        ######################################################
        # Set some attributes to their defaults, if necessary.
        # First we get the connection:
        if not connection and not getattr(cls, '_connection', None):
            mod = sys.modules[cls.__module__]
            # See if there's a __connection__ global in
            # the module, use it if there is.
            if hasattr(mod, '__connection__'):
                connection = mod.__connection__

        # Do not check hasattr(cls, '_connection') here - it is possible
        # SQLObject parent class has a connection attribute that came
        # from sqlhub, e.g.; check __dict__ only.
        if connection and ('_connection' not in cls.__dict__):
            cls.setConnection(connection)

        sqlmeta = cls.sqlmeta

        # We have to check if there are columns in the inherited
        # _columns where the attribute has been set to None in this
        # class.  If so, then we need to remove that column from
        # _columns.
        for key in sqlmeta.columnDefinitions.keys():
            if (key in new_attrs and new_attrs[key] is None):
                del sqlmeta.columnDefinitions[key]

        for column in sqlmeta.columnDefinitions.values():
            sqlmeta.addColumn(column)

        for column in implicitColumns:
            sqlmeta.addColumn(column)

        # Now the class is in an essentially OK-state, so we can
        # set up any magic attributes:
        declarative.setup_attributes(cls, new_attrs)

        if sqlmeta.fromDatabase:
            sqlmeta.addColumnsFromDatabase()

        for j in implicitJoins:
            sqlmeta.addJoin(j)
        for i in implicitIndexes:
            sqlmeta.addIndex(i)

        def order_getter(o):
            return o.creationOrder
        sqlmeta.columnList.sort(key=order_getter)
        sqlmeta.indexes.sort(key=order_getter)
        sqlmeta.indexDefinitions.sort(key=order_getter)
        # Joins cannot be sorted because addJoin created accessors
        # that remember indexes.
        # sqlmeta.joins.sort(key=order_getter)
        sqlmeta.joinDefinitions.sort(key=order_getter)

        # We don't setup the properties until we're finished with the
        # batch adding of all the columns...
        cls._notifyFinishClassCreation()
        cls._SO_finishedClassCreation = True
        makeProperties(cls)

        # We use the magic "q" attribute for accessing lazy
        # SQL where-clause generation.  See the sql module for
        # more.
        if not is_base:
            cls.q = sqlbuilder.SQLObjectTable(cls)
            cls.j = sqlbuilder.SQLObjectTableWithJoins(cls)

        classregistry.registry(sqlmeta.registry).addClass(cls)

    @classmethod
    def _SO_setupSqlmeta(cls, new_attrs, is_base):
        """
        This fixes up the sqlmeta attribute.  It handles both the case
        where no sqlmeta was given (in which we need to create another
        subclass), or the sqlmeta given doesn't have the proper
        inheritance.  Lastly it calls sqlmeta.setClass, which handles
        much of the setup.
        """
        if ('sqlmeta' not in new_attrs and not is_base):
            # We have to create our own subclass, usually.
            # type(className, bases_tuple, attr_dict) creates a new subclass.
            cls.sqlmeta = type('sqlmeta', (cls.sqlmeta,), {})
        if not issubclass(cls.sqlmeta, sqlmeta):
            # We allow no superclass and an object superclass, instead
            # of inheriting from sqlmeta; but in that case we replace
            # the class and just move over its attributes:
            assert cls.sqlmeta.__bases__ in ((), (object,)), (
                "If you do not inherit your sqlmeta class from "
                "sqlobject.sqlmeta, it must not inherit from any other "
                "class (your sqlmeta inherits from: %s)"
                % cls.sqlmeta.__bases__)
            for base in cls.__bases__:
                superclass = getattr(base, 'sqlmeta', None)
                if superclass:
                    break
            else:
                assert 0, (
                    "No sqlmeta class could be found in any superclass "
                    "(while fixing up sqlmeta %r inheritance)"
                    % cls.sqlmeta)
            values = dict(cls.sqlmeta.__dict__)
            for key in list(values.keys()):
                if key.startswith('__') and key.endswith('__'):
                    # Magic values shouldn't be passed through:
                    del values[key]
            cls.sqlmeta = type('sqlmeta', (superclass,), values)

        if not is_base:  # Do not pollute the base sqlmeta class
            cls.sqlmeta.setClass(cls)

    @classmethod
    def _SO_cleanDeprecatedAttrs(cls, new_attrs):
        """
        This removes attributes on SQLObject subclasses that have
        been deprecated; they are moved to the sqlmeta class, and
        a deprecation warning is given.
        """
        for attr in ():
            if attr in new_attrs:
                deprecated("%r is deprecated and read-only; please do "
                           "not use it in your classes until it is fully "
                           "deprecated" % attr, level=1, stacklevel=5)

    @classmethod
    def get(cls, id, connection=None, selectResults=None):

        assert id is not None, \
            'None is not a possible id for %s' % cls.__name__

        id = cls.sqlmeta.idType(id)

        if connection is None:
            cache = cls._connection.cache
        else:
            cache = connection.cache

        # This whole sequence comes from Cache.CacheFactory's
        # behavior, where a None returned means a cache miss.
        val = cache.get(id, cls)
        if val is None:
            try:
                val = cls(_SO_fetch_no_create=1)
                val._SO_validatorState = sqlbuilder.SQLObjectState(val)
                val._init(id, connection, selectResults)
                cache.put(id, cls, val)
            finally:
                cache.finishPut(cls)
        elif selectResults and not val.sqlmeta.dirty:
            val._SO_writeLock.acquire()
            try:
                val._SO_selectInit(selectResults)
                val.sqlmeta.expired = False
            finally:
                val._SO_writeLock.release()
        return val

    @classmethod
    def _notifyFinishClassCreation(cls):
        pass

    def _init(self, id, connection=None, selectResults=None):
        assert id is not None
        # This function gets called only when the object is
        # created, unlike __init__ which would be called
        # anytime the object was returned from cache.
        self.id = id
        self._SO_writeLock = threading.Lock()

        # If no connection was given, we'll inherit the class
        # instance variable which should have a _connection
        # attribute.
        if (connection is not None) and \
                (getattr(self, '_connection', None) is not connection):
            self._connection = connection
            # Sometimes we need to know if this instance is
            # global or tied to a particular connection.
            # This flag tells us that:
            self.sqlmeta._perConnection = True

        if not selectResults:
            dbNames = [col.dbName for col in self.sqlmeta.columnList]
            selectResults = self._connection._SO_selectOne(self, dbNames)
            if not selectResults:
                raise SQLObjectNotFound(
                    "The object %s by the ID %s does not exist" % (
                        self.__class__.__name__, self.id))
        self._SO_selectInit(selectResults)
        self._SO_createValues = {}
        self.sqlmeta.dirty = False

    def _SO_loadValue(self, attrName):
        try:
            return getattr(self, attrName)
        except AttributeError:
            try:
                self._SO_writeLock.acquire()
                try:
                    # Maybe, just in the moment since we got the lock,
                    # some other thread did a _SO_loadValue and we
                    # have the attribute!  Let's try and find out!  We
                    # can keep trying this all day and still beat the
                    # performance on the database call (okay, we can
                    # keep trying this for a few msecs at least)...
                    result = getattr(self, attrName)
                except AttributeError:
                    pass
                else:
                    return result
                self.sqlmeta.expired = False
                dbNames = [col.dbName for col in self.sqlmeta.columnList]
                selectResults = self._connection._SO_selectOne(self, dbNames)
                if not selectResults:
                    raise SQLObjectNotFound(
                        "The object %s by the ID %s has been deleted" % (
                            self.__class__.__name__, self.id))
                self._SO_selectInit(selectResults)
                result = getattr(self, attrName)
                return result
            finally:
                self._SO_writeLock.release()

    def sync(self):
        if self.sqlmeta.lazyUpdate and self._SO_createValues:
            self.syncUpdate()
        self._SO_writeLock.acquire()
        try:
            dbNames = [col.dbName for col in self.sqlmeta.columnList]
            selectResults = self._connection._SO_selectOne(self, dbNames)
            if not selectResults:
                raise SQLObjectNotFound(
                    "The object %s by the ID %s has been deleted" % (
                        self.__class__.__name__, self.id))
            self._SO_selectInit(selectResults)
            self.sqlmeta.expired = False
        finally:
            self._SO_writeLock.release()

    def syncUpdate(self):
        if not self._SO_createValues:
            return
        self._SO_writeLock.acquire()
        try:
            if self.sqlmeta.columns:
                columns = self.sqlmeta.columns
                values = [(columns[v[0]].dbName, v[1])
                          for v in sorted(
                              self._SO_createValues.items(),
                              key=lambda c: columns[c[0]].creationOrder)]
                self._connection._SO_update(self, values)
            self.sqlmeta.dirty = False
            self._SO_createValues = {}
        finally:
            self._SO_writeLock.release()

        post_funcs = []
        self.sqlmeta.send(events.RowUpdatedSignal, self, post_funcs)
        for func in post_funcs:
            func(self)

    def expire(self):
        if self.sqlmeta.expired:
            return
        self._SO_writeLock.acquire()
        try:
            if self.sqlmeta.expired:
                return
            for column in self.sqlmeta.columnList:
                delattr(self, instanceName(column.name))
            self.sqlmeta.expired = True
            self._connection.cache.expire(self.id, self.__class__)
            self._SO_createValues = {}
        finally:
            self._SO_writeLock.release()

    def _SO_setValue(self, name, value, from_python, to_python):
        # This is the place where we actually update the
        # database.

        # If we are _creating, the object doesn't yet exist
        # in the database, and we can't insert it until all
        # the parts are set.  So we just keep them in a
        # dictionary until later:
        d = {name: value}
        if not self.sqlmeta._creating and \
                not getattr(self.sqlmeta, "row_update_sig_suppress", False):
            self.sqlmeta.send(events.RowUpdateSignal, self, d)
        if len(d) != 1 or name not in d:
            # Already called RowUpdateSignal, don't call it again
            # inside .set()
            self.sqlmeta.row_update_sig_suppress = True
            self.set(**d)
            del self.sqlmeta.row_update_sig_suppress
        value = d[name]
        if from_python:
            dbValue = from_python(value, self._SO_validatorState)
        else:
            dbValue = value
        if to_python:
            value = to_python(dbValue, self._SO_validatorState)
        if self.sqlmeta._creating or self.sqlmeta.lazyUpdate:
            self.sqlmeta.dirty = True
            self._SO_createValues[name] = dbValue
            setattr(self, instanceName(name), value)
            return

        self._connection._SO_update(
            self, [(self.sqlmeta.columns[name].dbName,
                    dbValue)])

        if self.sqlmeta.cacheValues:
            setattr(self, instanceName(name), value)

        post_funcs = []
        self.sqlmeta.send(events.RowUpdatedSignal, self, post_funcs)
        for func in post_funcs:
            func(self)

    def set(self, _suppress_set_sig=False, **kw):
        if not self.sqlmeta._creating and \
                not getattr(self.sqlmeta, "row_update_sig_suppress", False) \
                and not _suppress_set_sig:
            self.sqlmeta.send(events.RowUpdateSignal, self, kw)
        # set() is used to update multiple values at once,
        # potentially with one SQL statement if possible.

        # Filter out items that don't map to column names.
        # Those will be set directly on the object using
        # setattr(obj, name, value).
        def is_column(_c):
            return _c in self.sqlmeta._plainSetters

        def f_is_column(item):
            return is_column(item[0])

        def f_not_column(item):
            return not is_column(item[0])
        items = kw.items()
        extra = dict(filter(f_not_column, items))
        kw = dict(filter(f_is_column, items))

        # _creating is special, see _SO_setValue
        if self.sqlmeta._creating or self.sqlmeta.lazyUpdate:
            for name, value in kw.items():
                from_python = getattr(self, '_SO_from_python_%s' % name, None)
                if from_python:
                    kw[name] = dbValue = from_python(value,
                                                     self._SO_validatorState)
                else:
                    dbValue = value
                to_python = getattr(self, '_SO_to_python_%s' % name, None)
                if to_python:
                    value = to_python(dbValue, self._SO_validatorState)
                setattr(self, instanceName(name), value)

            self._SO_createValues.update(kw)

            for name, value in extra.items():
                try:
                    getattr(self.__class__, name)
                except AttributeError:
                    if name not in self.sqlmeta.columns:
                        raise TypeError(
                            "%s.set() got an unexpected keyword argument "
                            "%s" % (self.__class__.__name__, name))
                try:
                    setattr(self, name, value)
                except AttributeError as e:
                    raise AttributeError('%s (with attribute %r)' % (e, name))

            self.sqlmeta.dirty = True
            return

        self._SO_writeLock.acquire()

        try:
            # We have to go through and see if the setters are
            # "plain", that is, if the user has changed their
            # definition in any way (put in something that
            # normalizes the value or checks for consistency,
            # for instance).  If so then we have to use plain
            # old setattr() to change the value, since we can't
            # read the user's mind.  We'll combine everything
            # else into a single UPDATE, if necessary.
            toUpdate = {}
            for name, value in kw.items():
                from_python = getattr(self, '_SO_from_python_%s' % name, None)
                if from_python:
                    dbValue = from_python(value, self._SO_validatorState)
                else:
                    dbValue = value
                to_python = getattr(self, '_SO_to_python_%s' % name, None)
                if to_python:
                    value = to_python(dbValue, self._SO_validatorState)
                if self.sqlmeta.cacheValues:
                    setattr(self, instanceName(name), value)
                toUpdate[name] = dbValue
            for name, value in extra.items():
                try:
                    getattr(self.__class__, name)
                except AttributeError:
                    if name not in self.sqlmeta.columns:
                        raise TypeError(
                            "%s.set() got an unexpected keyword argument "
                            "%s" % (self.__class__.__name__, name))
                try:
                    setattr(self, name, value)
                except AttributeError as e:
                    raise AttributeError('%s (with attribute %r)' % (e, name))

            if toUpdate:
                toUpdate = sorted(
                    toUpdate.items(),
                    key=lambda c: self.sqlmeta.columns[c[0]].creationOrder)
                args = [(self.sqlmeta.columns[name].dbName, value)
                        for name, value in toUpdate]
                self._connection._SO_update(self, args)
        finally:
            self._SO_writeLock.release()

        post_funcs = []
        self.sqlmeta.send(events.RowUpdatedSignal, self, post_funcs)
        for func in post_funcs:
            func(self)

    def _SO_selectInit(self, row):
        for _col, colValue in zip(self.sqlmeta.columnList, row):
            if _col.to_python:
                colValue = _col.to_python(colValue, self._SO_validatorState)
            setattr(self, instanceName(_col.name), colValue)

    def _SO_getValue(self, name):
        # Retrieves a single value from the database.  Simple.
        assert not self.sqlmeta._obsolete, (
            "%s with id %s has become obsolete"
            % (self.__class__.__name__, self.id))
        # @@: do we really need this lock?
        # self._SO_writeLock.acquire()
        column = self.sqlmeta.columns[name]
        results = self._connection._SO_selectOne(self, [column.dbName])
        # self._SO_writeLock.release()
        assert results is not None, "%s with id %s is not in the database" % (
            self.__class__.__name__, self.id)
        value = results[0]
        if column.to_python:
            value = column.to_python(value, self._SO_validatorState)
        return value

    def _SO_foreignKey(self, value, joinClass, idName=None):
        if value is None:
            return None
        if self.sqlmeta._perConnection:
            connection = self._connection
        else:
            connection = None
        if idName is None:  # Get by id
            return joinClass.get(value, connection=connection)
        return joinClass.select(
            getattr(joinClass.q, idName) == value,
            connection=connection).getOne()

    def __init__(self, **kw):
        # If we are the outmost constructor of a hiearchy of
        # InheritableSQLObjects (or simlpy _the_ constructor of a "normal"
        # SQLObject), we create a threadlocal list that collects the
        # RowCreatedSignals, and executes them if this very constructor is left
        try:
            _postponed_local.postponed_calls
            postponed_created = False
        except AttributeError:
            _postponed_local.postponed_calls = []
            postponed_created = True

        try:
            # We shadow the sqlmeta class with an instance of sqlmeta
            # that points to us (our sqlmeta buddy object; where the
            # sqlmeta class is our class's buddy class)
            self.sqlmeta = self.__class__.sqlmeta(self)
            # The get() classmethod/constructor uses a magic keyword
            # argument when it wants an empty object, fetched from the
            # database.  So we have nothing more to do in that case:
            if '_SO_fetch_no_create' in kw:
                return

            post_funcs = []
            self.sqlmeta.send(events.RowCreateSignal, self, kw, post_funcs)

            # Pass the connection object along if we were given one.
            if 'connection' in kw:
                connection = kw.pop('connection')
                if getattr(self, '_connection', None) is not connection:
                    self._connection = connection
                    self.sqlmeta._perConnection = True

            self._SO_writeLock = threading.Lock()

            if 'id' in kw:
                id = self.sqlmeta.idType(kw['id'])
                del kw['id']
            else:
                id = None

            self._create(id, **kw)

            for func in post_funcs:
                func(self)
        finally:
            # if we are the creator of the tl-storage, we
            # have to exectute and under all circumstances
            # remove the tl-storage
            if postponed_created:
                try:
                    for func in _postponed_local.postponed_calls:
                        func()
                finally:
                    del _postponed_local.postponed_calls

    def _create(self, id, **kw):

        self.sqlmeta._creating = True
        self._SO_createValues = {}
        self._SO_validatorState = sqlbuilder.SQLObjectState(self)

        # First we do a little fix-up on the keywords we were
        # passed:
        for column in self.sqlmeta.columnList:

            # Then we check if the column wasn't passed in, and
            # if not we try to get the default.
            if column.name not in kw and column.foreignName not in kw:
                default = column.default

                # If we don't get it, it's an error:
                # If we specified an SQL DEFAULT, then we should use that
                if default is NoDefault:
                    if column.defaultSQL is None:
                        raise TypeError(
                            "%s() did not get expected keyword argument "
                            "'%s'" % (self.__class__.__name__, column.name))
                    else:
                        # There is defaultSQL for the column -
                        # do not put the column to kw
                        # so that the backend creates the value.
                        continue

                # Otherwise we put it in as though they did pass
                # that keyword:

                kw[column.name] = default

        self.set(**kw)

        # Then we finalize the process:
        self._SO_finishCreate(id)

    def _SO_finishCreate(self, id=None):
        # Here's where an INSERT is finalized.
        # These are all the column values that were supposed
        # to be set, but were delayed until now:
        setters = self._SO_createValues.items()
        setters = sorted(
            setters, key=lambda c: self.sqlmeta.columns[c[0]].creationOrder)
        # Here's their database names:
        names = [self.sqlmeta.columns[v[0]].dbName for v in setters]
        values = [v[1] for v in setters]
        # Get rid of _SO_create*, we aren't creating anymore.
        # Doesn't have to be threadsafe because we're still in
        # new(), which doesn't need to be threadsafe.
        self.sqlmeta.dirty = False
        if not self.sqlmeta.lazyUpdate:
            del self._SO_createValues
        else:
            self._SO_createValues = {}
        del self.sqlmeta._creating

        # Do the insert -- most of the SQL in this case is left
        # up to DBConnection, since getting a new ID is
        # non-standard.
        id = self._connection.queryInsertID(self,
                                            id, names, values)
        cache = self._connection.cache
        cache.created(id, self.__class__, self)
        self._init(id)
        post_funcs = []
        kw = dict([('class', self.__class__), ('id', id)])

        def _send_RowCreatedSignal():
            self.sqlmeta.send(events.RowCreatedSignal, self, kw, post_funcs)
            for func in post_funcs:
                func(self)
        _postponed_local.postponed_calls.append(_send_RowCreatedSignal)

    def _SO_getID(self, obj, refColumn=None):
        return getID(obj, refColumn)

    @classmethod
    def _findAlternateID(cls, name, dbName, value, connection=None):
        if isinstance(name, str):
            name = (name,)
            value = (value,)
        if len(name) != len(value):
            raise ValueError(
                "'column' and 'value' tuples must be of the same size")
        new_value = []
        for n, v in zip(name, value):
            from_python = getattr(cls, '_SO_from_python_' + n)
            if from_python:
                v = from_python(
                    v, sqlbuilder.SQLObjectState(cls, connection=connection))
            new_value.append(v)
        condition = sqlbuilder.AND(
            *[getattr(cls.q, _n) == _v for _n, _v in zip(name, new_value)])
        return (connection or cls._connection)._SO_selectOneAlt(
            cls,
            [cls.sqlmeta.idName]
            + [column.dbName for column in cls.sqlmeta.columnList],
            condition), None

    @classmethod
    def _SO_fetchAlternateID(cls, name, dbName, value, connection=None,
                             idxName=None):
        result, obj = cls._findAlternateID(name, dbName, value, connection)
        if not result:
            if idxName is None:
                raise SQLObjectNotFound(
                    "The %s by alternateID %s = %s does not exist" % (
                        cls.__name__, name, repr(value)))
            else:
                names = []
                for i in range(len(name)):
                    names.append("%s = %s" % (name[i], repr(value[i])))
                names = ', '.join(names)
                raise SQLObjectNotFound(
                    "The %s by unique index %s(%s) does not exist" % (
                        cls.__name__, idxName, names))
        if obj:
            return obj
        if connection:
            obj = cls.get(result[0], connection=connection,
                          selectResults=result[1:])
        else:
            obj = cls.get(result[0], selectResults=result[1:])
        return obj

    @classmethod
    def _SO_depends(cls):
        return findDependencies(cls.__name__, cls.sqlmeta.registry)

    @classmethod
    def select(cls, clause=None, clauseTables=None,
               orderBy=NoDefault, limit=None,
               lazyColumns=False, reversed=False,
               distinct=False, connection=None,
               join=None, forUpdate=False):
        return cls.SelectResultsClass(cls, clause,
                                      clauseTables=clauseTables,
                                      orderBy=orderBy,
                                      limit=limit,
                                      lazyColumns=lazyColumns,
                                      reversed=reversed,
                                      distinct=distinct,
                                      connection=connection,
                                      join=join, forUpdate=forUpdate)

    @classmethod
    def selectBy(cls, connection=None, **kw):
        conn = connection or cls._connection
        return cls.SelectResultsClass(cls,
                                      conn._SO_columnClause(cls, kw),
                                      connection=conn)

    @classmethod
    def tableExists(cls, connection=None):
        conn = connection or cls._connection
        return conn.tableExists(cls.sqlmeta.table)

    @classmethod
    def dropTable(cls, ifExists=False, dropJoinTables=True, cascade=False,
                  connection=None):
        conn = connection or cls._connection
        if ifExists and not cls.tableExists(connection=conn):
            return
        extra_sql = []
        post_funcs = []
        cls.sqlmeta.send(events.DropTableSignal, cls, connection,
                         extra_sql, post_funcs)
        conn.dropTable(cls.sqlmeta.table, cascade)
        if dropJoinTables:
            cls.dropJoinTables(ifExists=ifExists, connection=conn)
        for sql in extra_sql:
            connection.query(sql)
        for func in post_funcs:
            func(cls, conn)

    @classmethod
    def createTable(cls, ifNotExists=False, createJoinTables=True,
                    createIndexes=True, applyConstraints=True,
                    connection=None):
        conn = connection or cls._connection
        if ifNotExists and cls.tableExists(connection=conn):
            return
        extra_sql = []
        post_funcs = []
        cls.sqlmeta.send(events.CreateTableSignal, cls, connection,
                         extra_sql, post_funcs)
        constraints = conn.createTable(cls)
        if applyConstraints:
            for constraint in constraints:
                conn.query(constraint)
        else:
            extra_sql.extend(constraints)
        if createJoinTables:
            cls.createJoinTables(ifNotExists=ifNotExists,
                                 connection=conn)
        if createIndexes:
            cls.createIndexes(ifNotExists=ifNotExists,
                              connection=conn)
        for func in post_funcs:
            func(cls, conn)
        return extra_sql

    @classmethod
    def createTableSQL(cls, createJoinTables=True, createIndexes=True,
                       connection=None):
        conn = connection or cls._connection
        sql, constraints = conn.createTableSQL(cls)
        if createJoinTables:
            join_sql = cls.createJoinTablesSQL(connection=conn)
            if join_sql:
                sql += ';\n' + join_sql
        if createIndexes:
            index_sql = cls.createIndexesSQL(connection=conn)
            if index_sql:
                sql += ';\n' + index_sql
        return sql, constraints

    @classmethod
    def createJoinTables(cls, ifNotExists=False, connection=None):
        conn = connection or cls._connection
        for join in cls._getJoinsToCreate():
            if (ifNotExists
                    and conn.tableExists(join.intermediateTable)):
                continue
            conn._SO_createJoinTable(join)

    @classmethod
    def createJoinTablesSQL(cls, connection=None):
        conn = connection or cls._connection
        sql = []
        for join in cls._getJoinsToCreate():
            sql.append(conn._SO_createJoinTableSQL(join))
        return ';\n'.join(sql)

    @classmethod
    def createIndexes(cls, ifNotExists=False, connection=None):
        conn = connection or cls._connection
        for _index in cls.sqlmeta.indexes:
            if not _index:
                continue
            conn._SO_createIndex(cls, _index)

    @classmethod
    def createIndexesSQL(cls, connection=None):
        conn = connection or cls._connection
        sql = []
        for _index in cls.sqlmeta.indexes:
            if not _index:
                continue
            sql.append(conn.createIndexSQL(cls, _index))
        return ';\n'.join(sql)

    @classmethod
    def _getJoinsToCreate(cls):
        joins = []
        for join in cls.sqlmeta.joins:
            if not join:
                continue
            if not join.hasIntermediateTable() or \
                    not getattr(join, 'createRelatedTable', True):
                continue
            if join.soClass.__name__ > join.otherClass.__name__:
                continue
            joins.append(join)
        return joins

    @classmethod
    def dropJoinTables(cls, ifExists=False, connection=None):
        conn = connection or cls._connection
        for join in cls.sqlmeta.joins:
            if not join:
                continue
            if not join.hasIntermediateTable() or \
                    not getattr(join, 'createRelatedTable', True):
                continue
            if join.soClass.__name__ > join.otherClass.__name__:
                continue
            if ifExists and \
               not conn.tableExists(join.intermediateTable):
                continue
            conn._SO_dropJoinTable(join)

    @classmethod
    def clearTable(cls, connection=None, clearJoinTables=True):
        # 3-03 @@: Maybe this should check the cache... but it's
        # kind of crude anyway, so...
        conn = connection or cls._connection
        conn.clearTable(cls.sqlmeta.table)
        if clearJoinTables:
            for join in cls._getJoinsToCreate():
                conn.clearTable(join.intermediateTable)

    def destroySelf(self):
        post_funcs = []
        self.sqlmeta.send(events.RowDestroySignal, self, post_funcs)
        # Kills this object.  Kills it dead!

        klass = self.__class__

        # Free related joins on the base class
        for join in klass.sqlmeta.joins:
            if isinstance(join, joins.SORelatedJoin):
                q = "DELETE FROM %s WHERE %s=%d" % (join.intermediateTable,
                                                    join.joinColumn, self.id)
                self._connection.query(q)

        depends = self._SO_depends()
        for k in depends:
            # Free related joins
            for join in k.sqlmeta.joins:
                if isinstance(join, joins.SORelatedJoin) and \
                        join.otherClassName == klass.__name__:
                    q = "DELETE FROM %s WHERE %s=%d" % (join.intermediateTable,
                                                        join.otherColumn,
                                                        self.id)
                    self._connection.query(q)

            cols = findDependantColumns(klass.__name__, k)

            # Don't confuse the rest of the process
            if len(cols) == 0:
                continue

            query = []
            restrict = False
            for _col in cols:
                query.append(getattr(k.q, _col.name) == self.id)
                if _col.cascade is False:
                    # Found a restriction
                    restrict = True
            query = sqlbuilder.OR(*query)
            results = k.select(query, connection=self._connection)
            if restrict and results.count():
                # Restrictions only apply if there are
                # matching records on the related table
                raise SQLObjectIntegrityError(
                    "Tried to delete %s::%s but "
                    "table %s has a restriction against it" %
                    (klass.__name__, self.id, k.__name__))

            setnull = {}
            for _col in cols:
                if _col.cascade == 'null':
                    setnull[_col.name] = None
            if setnull:
                for row in results:
                    clear = {}
                    for name in setnull:
                        if getattr(row, name) == self.id:
                            clear[name] = None
                    row.set(**clear)

            delete = False
            for _col in cols:
                if _col.cascade is True:
                    delete = True
            assert delete or setnull or restrict, (
                "Class %s depends on %s accoriding to "
                "findDependantColumns, but this seems inaccurate"
                % (k, klass))
            if delete:
                for row in results:
                    row.destroySelf()

        self.sqlmeta._obsolete = True
        self._connection._SO_delete(self)
        self._connection.cache.expire(self.id, self.__class__)

        for func in post_funcs:
            func(self)

        post_funcs = []
        self.sqlmeta.send(events.RowDestroyedSignal, self, post_funcs)
        for func in post_funcs:
            func(self)

    @classmethod
    def delete(cls, id, connection=None):
        obj = cls.get(id, connection=connection)
        obj.destroySelf()

    @classmethod
    def deleteMany(cls, where=NoDefault, connection=None):
        conn = connection or cls._connection
        conn.query(conn.sqlrepr(sqlbuilder.Delete(cls.sqlmeta.table, where)))

    @classmethod
    def deleteBy(cls, connection=None, **kw):
        conn = connection or cls._connection
        conn.query(conn.sqlrepr(sqlbuilder.Delete(
            cls.sqlmeta.table, conn._SO_columnClause(cls, kw))))

    def __repr__(self):
        if not hasattr(self, 'id'):
            # Object initialization not finished.  No attributes can be read.
            return '<%s (not initialized)>' % self.__class__.__name__
        return '<%s %r %s>' \
               % (self.__class__.__name__,
                  self.id,
                  ' '.join(
                      ['%s=%s' % (name, repr(value))
                       for name, value in self._reprItems()]))

    def __sqlrepr__(self, db):
        return str(self.id)

    @classmethod
    def sqlrepr(cls, value, connection=None):
        return (connection or cls._connection).sqlrepr(value)

    @classmethod
    def coerceID(cls, value):
        if isinstance(value, cls):
            return value.id
        else:
            return cls.sqlmeta.idType(value)

    def _reprItems(self):
        items = []
        for _col in self.sqlmeta.columnList:
            value = getattr(self, _col.name)
            r = repr(value)
            if len(r) > 20:
                value = r[:17] + "..." + r[-1]
            items.append((_col.name, value))
        return items

    @classmethod
    def setConnection(cls, value):
        if isinstance(value, string_type):
            value = dbconnection.connectionForURI(value)
        cls._connection = value

    def tablesUsedImmediate(self):
        return [self.__class__.q]

    # hash implementation

    def __hash__(self):
        # We hash on class name and id, since that should be
        # unique
        return hash((self.__class__.__name__, self.id))

    # Comparison

    def __eq__(self, other):
        if self.__class__ is other.__class__:
            if self.id == other.id:
                return True
        return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __lt__(self, other):
        return NotImplemented

    def __le__(self, other):
        return NotImplemented

    def __gt__(self, other):
        return NotImplemented

    def __ge__(self, other):
        return NotImplemented

    # (De)serialization (pickle, etc.)

    def __getstate__(self):
        if self.sqlmeta._perConnection:
            from pickle import PicklingError
            raise PicklingError(
                'Cannot pickle an SQLObject instance '
                'that has a per-instance connection')
        if self.sqlmeta.lazyUpdate and self._SO_createValues:
            self.syncUpdate()
        d = self.__dict__.copy()
        del d['sqlmeta']
        del d['_SO_validatorState']
        del d['_SO_writeLock']
        del d['_SO_createValues']
        return d

    def __setstate__(self, d):
        self.__init__(_SO_fetch_no_create=1)
        self._SO_validatorState = sqlbuilder.SQLObjectState(self)
        self._SO_writeLock = threading.Lock()
        self._SO_createValues = {}
        self.__dict__.update(d)
        cls = self.__class__
        cache = self._connection.cache
        if cache.tryGet(self.id, cls) is not None:
            raise ValueError(
                "Cannot unpickle %s row with id=%s - "
                "a different instance with the id already exists "
                "in the cache" % (cls.__name__, self.id))
        cache.created(self.id, cls, self)


def setterName(name):
    return '_set_%s' % name


def rawSetterName(name):
    return '_SO_set_%s' % name


def getterName(name):
    return '_get_%s' % name


def rawGetterName(name):
    return '_SO_get_%s' % name


def instanceName(name):
    return '_SO_val_%s' % name


########################################
# Utility functions (for external consumption)
########################################

def getID(obj, refColumn=None):
    if isinstance(obj, SQLObject):
        return getattr(obj, refColumn or 'id')
    elif isinstance(obj, int):
        return obj
    elif isinstance(obj, long):
        return int(obj)
    elif isinstance(obj, str):
        try:
            return int(obj)
        except ValueError:
            return obj
    elif obj is None:
        return None


def getObject(obj, klass):
    if isinstance(obj, int):
        return klass(obj)
    elif isinstance(obj, long):
        return klass(int(obj))
    elif isinstance(obj, str):
        return klass(int(obj))
    elif obj is None:
        return None
    else:
        return obj

__all__ = [
    'NoDefault', 'SQLObject', 'SQLObjectIntegrityError', 'SQLObjectNotFound',
    'getID', 'getObject', 'sqlhub', 'sqlmeta',
]



from hashlib import sha256, md5
from sqlobject import AND, IntCol, OR, SQLObject
from sqlobject.compat import PY2
from sqlobject.include import hashcol
from sqlobject.tests.dbtest import setupClass

########################################
# HashCol test
########################################

if PY2:
    def sha256_str(x):
        return sha256(x).hexdigest()

    def md5_str(x):
        return md5(x).hexdigest()
else:
    def sha256_str(x):
        return sha256(x.encode('utf8')).hexdigest()

    def md5_str(x):
        return md5(x.encode('utf8')).hexdigest()


class HashTest(SQLObject):
    so_count = IntCol(alternateID=True)
    col1 = hashcol.HashCol()
    col2 = hashcol.HashCol(hashMethod=sha256_str)


data = ['test', 'This is more text', 'test 2']
items = []


def setup():
    global items
    items = []
    setupClass(HashTest)
    for i, s in enumerate(data):
        items.append(HashTest(so_count=i, col1=s, col2=s))


def test_create():
    setup()
    for s, item in zip(data, items):
        assert item.col1 == s
        assert item.col2 == s

    conn = HashTest._connection
    rows = conn.queryAll("""
    SELECT so_count, col1, col2
    FROM hash_test
    ORDER BY so_count
    """)
    for so_count, col1, col2 in rows:
        assert md5_str(data[so_count]) == col1
        assert sha256_str(data[so_count]) == col2


def test_select():
    for i, value in enumerate(data):
        rows = list(HashTest.select(HashTest.q.col1 == value))
        assert len(rows) == 1
        rows = list(HashTest.select(HashTest.q.col2 == value))
        assert len(rows) == 1
        # Passing the hash in directly should fail
        rows = list(HashTest.select(HashTest.q.col2 == sha256_str(value)))
        assert len(rows) == 0
        rows = list(HashTest.select(AND(
            HashTest.q.col1 == value,
            HashTest.q.col2 == value
        )))
        assert len(rows) == 1
        rows = list(HashTest.selectBy(col1=value))
        assert len(rows) == 1
        rows = list(HashTest.selectBy(col2=value))
        assert len(rows) == 1
        rows = list(HashTest.selectBy(col1=value, col2=value))
        assert len(rows) == 1
    rows = list(HashTest.select(OR(
        HashTest.q.col1 == 'test 2',
        HashTest.q.col2 == 'test'
    )))
    assert len(rows) == 2



from sqlobject import ForeignKey, ManyToMany, OneToMany, SQLObject, StringCol
from .dbtest import setupClass


class ComplexGroup(SQLObject):
    name = StringCol()
    complexes = OneToMany('Complex')

    def _get_unit_models(self):
        q = self.complexes.clause & Complex.unit_models.clause
        return UnitModel.select(q)


class Complex(SQLObject):
    name = StringCol()
    unit_models = ManyToMany('UnitModel')
    complex_group = ForeignKey('ComplexGroup')


class UnitModel(SQLObject):
    class sqlmeta:
        defaultOrderBy = 'name'
    name = StringCol()
    complexes = ManyToMany('Complex')


def test_join_sqlrepr():
    setupClass([ComplexGroup, UnitModel, Complex])
    cg1 = ComplexGroup(name='cg1')
    cg2 = ComplexGroup(name='cg2')
    c1 = Complex(name='c1', complex_group=cg1)
    c2 = Complex(name='c2', complex_group=cg2)
    c3 = Complex(name='c3', complex_group=cg2)
    u1 = UnitModel(name='u1')
    u2 = UnitModel(name='u2')
    u1.complexes.add(c1)
    u1.complexes.add(c2)
    u2.complexes.add(c2)
    u2.complexes.add(c3)
    assert list(Complex.selectBy(name='c1')) == [c1]

    assert list(cg1.unit_models) == [u1]
    assert list(cg2.unit_models) == [u1, u2, u2]
    assert list(cg2.unit_models.distinct()) == [u1, u2]

    assert list(
        cg2.unit_models.filter(UnitModel.q.name == 'u1')) == [u1]



"""
Contributed by Edigram SAS, Paris France Tel:01 44 77 94 00
Ahmed MOHAMED ALI <ahmedmoali@yahoo.com> 27 April 2004

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

connection creation sample::

    __connection__ = DBConnection.maxdbConnection(
        host=hostname, database=dbname,
        user=user_name, password=user_password, autoCommit=1, debug=1)
"""

import os
from sqlobject.dbconnection import DBAPI
from sqlobject import col


class maxdbException(Exception):

    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


class LowerBoundOfSliceIsNotSupported(maxdbException):
    def __init__(self, value):
        maxdbException.__init__(self, '')


class IncorrectIDStyleError(maxdbException):
    def __init__(self, value):
        maxdbException.__init__(
            self,
            'This primary key name is not in the expected style, '
            'please rename the column to %r or switch to another style'
            % value)


class StyleMismatchError(maxdbException):
    def __init__(self, value):
        maxdbException.__init__(
            self,
            'The name %r is only permitted for primary key, change the '
            'column name or switch to another style' % value)


class PrimaryKeyNotFounded(maxdbException):
    def __init__(self, value):
        maxdbException.__init__(
            self,
            "No primary key was defined on table %r" % value)


SAPDBMAX_ID_LENGTH = 32


class MaxdbConnection(DBAPI):

    supportTransactions = True
    dbName = 'maxdb'
    schemes = [dbName]

    def __init__(self, host='', port=None, user=None, password=None,
                 database=None, autoCommit=1, sqlmode='internal',
                 isolation=None, timeout=None, **kw):
        from sapdb import dbapi
        self.module = dbapi
        self.host = host
        self.port = port
        self.user = user
        self.password = password
        self.db = database
        self.autoCommit = autoCommit
        self.sqlmode = sqlmode
        self.isolation = isolation
        self.timeout = timeout

        DBAPI.__init__(self, **kw)

    @classmethod
    def _connectionFromParams(cls, auth, password, host, port, path, args):
        path = path.replace('/', os.path.sep)
        return cls(host, port, user=auth, password=password, database=path,
                   **args)

    def _getConfigParams(self, sqlmode, auto):
        autocommit = 'off'
        if auto:
            autocommit = 'on'
        opt = {}
        opt["autocommit"] = autocommit
        opt["sqlmode"] = sqlmode
        if self.isolation:
            opt["isolation"] = self.isolation
        if self.timeout:
            opt["timeout"] = self.timeout
        return opt

    def _setAutoCommit(self, conn, auto):
        conn.close()
        conn.__init__(self.user, self.password, self.db, self.host,
                      **self._getConfigParams(self.sqlmode, auto))

    def createSequenceName(self, table):
        """
        sequence name are builded with the concatenation of the table
        name with '_SEQ' word we truncate the name of the
        sequence_name because sapdb identifier cannot exceed 32
        characters so that the name of the sequence does not exceed 32
        characters
        """
        return '%s_SEQ' % (table[:SAPDBMAX_ID_LENGTH - 4])

    def makeConnection(self):
        conn = self.module.Connection(
            self.user, self.password, self.db, self.host,
            **self._getConfigParams(self.sqlmode, self.autoCommit))
        return conn

    def _queryInsertID(self, conn, soInstance, id, names, values):
        table = soInstance.sqlmeta.table
        idName = soInstance.sqlmeta.idName
        c = conn.cursor()
        if id is None:
            c.execute(
                'SELECT %s.NEXTVAL FROM DUAL' % (
                    self.createSequenceName(table)))
            id = c.fetchone()[0]
        names = [idName] + names
        values = [id] + values
        q = self._insertSQL(table, names, values)
        if self.debug:
            self.printDebug(conn, q, 'QueryIns')
        c.execute(q)
        c.close()
        if self.debugOutput:
            self.printDebug(conn, id, 'QueryIns', 'result')
        return id

    @classmethod
    def sqlAddLimit(cls, query, limit):
        sql = query
        sql = sql.replace("SELECT", "SELECT ROWNO, ")
        if sql.find('WHERE') != -1:
            sql = sql + ' AND ' + limit
        else:
            sql = sql + 'WHERE ' + limit
        return sql

    @classmethod
    def _queryAddLimitOffset(cls, query, start, end):
        if start:
            raise LowerBoundOfSliceIsNotSupported
        limit = ' ROWNO   <= %d ' % (end)
        return cls.sqlAddLimit(query, limit)

    def createTable(self, soClass):
        # We create the table in a transaction because the addition of the
        # table and the sequence must be atomic.

        # I tried to use the transaction class
        # but I get a recursion limit error.
        # t=self.transaction()
        # t.query('CREATE TABLE %s (\n%s\n)' %
        #            (soClass.sqlmeta.table, self.createColumns(soClass)))
        #
        # t.query("CREATE SEQUENCE %s" %
        #         self.createSequenceName(soClass.sqlmeta.table))
        # t.commit()
        # so use transaction when the problem will be solved
        self.query('CREATE TABLE %s (\n%s\n)' %
                   (soClass.sqlmeta.table, self.createColumns(soClass)))
        self.query("CREATE SEQUENCE %s"
                   % self.createSequenceName(soClass.sqlmeta.table))
        return []

    def createReferenceConstraint(self, soClass, col):
        return col.maxdbCreateReferenceConstraint()

    def createColumn(self, soClass, col):
        return col.maxdbCreateSQL()

    def createIDColumn(self, soClass):
        key_type = {int: "INT", str: "TEXT"}[soClass.sqlmeta.idType]
        return '%s %s PRIMARY KEY' % (soClass.sqlmeta.idName, key_type)

    def createIndexSQL(self, soClass, index):
        return index.maxdbCreateIndexSQL(soClass)

    def dropTable(self, tableName, cascade=False):
        # We drop the table in a transaction because the removal of the
        # table and the sequence must be atomic.
        # I tried to use the transaction class
        # but I get a recursion limit error.
        # try:
        #     t=self.transaction()
        #     t.query("DROP TABLE %s" % tableName)
        #     t.query("DROP SEQUENCE %s" % self.createSequenceName(tableName))
        #     t.commit()
        # except:
        #     t.rollback()
        # so use transaction when the problem will be solved
        self.query("DROP TABLE %s" % tableName)
        self.query("DROP SEQUENCE %s" % self.createSequenceName(tableName))

    def joinSQLType(self, join):
        return 'INT NOT NULL'

    def tableExists(self, tableName):
        for (table,) in self.queryAll(
                "SELECT OBJECT_NAME FROM ALL_OBJECTS "
                "WHERE OBJECT_TYPE='TABLE'"):
            if table.lower() == tableName.lower():
                return True
        return False

    def addColumn(self, tableName, column):
        self.query('ALTER TABLE %s ADD %s' %
                   (tableName,
                    column.maxdbCreateSQL()))

    def delColumn(self, sqlmeta, column):
        self.query('ALTER TABLE %s DROP COLUMN %s' % (sqlmeta.table,
                                                      column.dbName))

    GET_COLUMNS = """
    SELECT COLUMN_NAME, NULLABLE, DATA_DEFAULT, DATA_TYPE,
           DATA_LENGTH, DATA_SCALE
    FROM USER_TAB_COLUMNS WHERE TABLE_NAME=UPPER('%s')"""

    GET_PK_AND_FK = """
    SELECT constraint_cols.column_name, constraints.constraint_type,
           refname,reftablename
    FROM user_cons_columns constraint_cols
    INNER JOIN user_constraints constraints
    ON constraint_cols.constraint_name = constraints.constraint_name
    LEFT OUTER JOIN show_foreign_key fk
    ON constraint_cols.column_name = fk.columnname
    WHERE constraints.table_name =UPPER('%s')"""

    def columnsFromSchema(self, tableName, soClass):
        colData = self.queryAll(self.GET_COLUMNS
                                % tableName)

        results = []
        keymap = {}
        pkmap = {}
        fkData = self.queryAll(self.GET_PK_AND_FK % tableName)
        for _col, cons_type, refcol, reftable in fkData:
            col_name = _col.lower()
            pkmap[col_name] = False
            if cons_type == 'R':
                keymap[col_name] = reftable.lower()

            elif cons_type == 'P':
                pkmap[col_name] = True

        if len(pkmap) == 0:
            raise PrimaryKeyNotFounded(tableName)

        for (field, nullAllowed, default, data_type, data_len,
             data_scale) in colData:
            # id is defined as primary key --> ok
            # We let sqlobject raise error if the 'id' is used
            # for another column.
            field_name = field.lower()
            if (field_name == soClass.sqlmeta.idName) and pkmap[field_name]:
                continue

            colClass, kw = self.guessClass(data_type, data_len, data_scale)
            kw['name'] = field_name
            kw['dbName'] = field

            if nullAllowed == 'Y':
                nullAllowed = False
            else:
                nullAllowed = True

            kw['notNone'] = nullAllowed
            if default is not None:
                kw['default'] = default

            if field_name in keymap:
                kw['foreignKey'] = keymap[field_name]

            results.append(colClass(**kw))

        return results

    _numericTypes = ['INTEGER', 'INT', 'SMALLINT']
    _dateTypes = ['DATE', 'TIME', 'TIMESTAMP']

    def guessClass(self, t, flength, fscale=None):
        """
        An internal method that tries to figure out what Col subclass
        is appropriate given whatever introspective information is
        available -- both very database-specific.
        """
        if t in self._numericTypes:
            return col.IntCol, {}
        # The type returned by the sapdb library for LONG is
        # SapDB_LongReader To get the data call the read member with
        # desired size (default =-1 means get all)

        elif t.find('LONG') != -1:
            return col.StringCol, {'length': flength,
                                   'varchar': False}
        elif t in self._dateTypes:
            return col.DateTimeCol, {}
        elif t == 'FIXED':
            return col.CurrencyCol, {'size': flength,
                                     'precision': fscale}
        else:
            return col.Col, {}



from __future__ import print_function

from sqlobject import BoolCol, SQLObject
from sqlobject.sqlbuilder import AND, Alias, EXISTS, JOIN, LEFTJOINOn, \
    Select, sqlrepr
from sqlobject.tests.dbtest import setupClass

''' Going to test that complex sqlbuilder constructions are never
    prematurely stringified. A straight-forward approach is to use
    Bools, since postgresql wants special formatting in queries.
    The test is whether a call to sqlrepr(x, 'postgres') includes
    the appropriate bool formatting throughout.
'''


class SBButton(SQLObject):
    activated = BoolCol()


def makeClause():
    # It's not a comparison, it's an SQLExpression
    return SBButton.q.activated == True  # noqa


def makeSelect():
    return Select(SBButton.q.id, clause=makeClause())


def checkCount(q, c, msg=''):
    print("STRING:", str(q))
    print("POSTGR:", sqlrepr(q, 'postgres'))
    assert sqlrepr(q, 'postgres').count("'t'") == c and \
        sqlrepr(q, 'postgres') != str(q), msg


def testSimple():
    setupClass(SBButton)
    checkCount(makeClause(), 1)
    checkCount(makeSelect(), 1)


def testMiscOps():
    setupClass(SBButton)
    checkCount(AND(makeClause(), makeClause()), 2)
    checkCount(AND(makeClause(), EXISTS(makeSelect())), 2)


def testAliased():
    setupClass(SBButton)
    b = Alias(makeSelect(), 'b')
    checkCount(b, 1)
    checkCount(Select(b.q.id), 1)

    # Table1 & Table2 are treated individually in joins
    checkCount(JOIN(None, b), 1)
    checkCount(JOIN(b, SBButton), 1)
    checkCount(JOIN(SBButton, b), 1)
    checkCount(LEFTJOINOn(None, b, SBButton.q.id == b.q.id), 1)
    checkCount(LEFTJOINOn(b, SBButton, SBButton.q.id == b.q.id), 1)
    checkCount(LEFTJOINOn(SBButton, b, SBButton.q.id == b.q.id), 1)


def testTablesUsedSResults():
    setupClass(SBButton)

    checkCount(SBButton.select(makeClause()).queryForSelect(), 1)



from sqlobject import ForeignKey, SQLObject, StringCol
from sqlobject.sqlbuilder import EXISTS, IN, LEFTOUTERJOINOn, NOTEXISTS, \
    Outer, Select
from sqlobject.tests.dbtest import setupClass


########################################
# Subqueries (subselects)
########################################


class SOTestIn1(SQLObject):
    col1 = StringCol()


class SOTestIn2(SQLObject):
    col2 = StringCol()


class SOTestOuter(SQLObject):
    fk = ForeignKey('SOTestIn1')


def setup():
    setupClass(SOTestIn1)
    setupClass(SOTestIn2)


def insert():
    setup()
    SOTestIn1(col1=None)
    SOTestIn1(col1='')
    SOTestIn1(col1="test")
    SOTestIn2(col2=None)
    SOTestIn2(col2='')
    SOTestIn2(col2="test")


def test_1syntax_in():
    setup()
    select = SOTestIn1.select(IN(SOTestIn1.q.col1, Select(SOTestIn2.q.col2)))
    assert str(select) == \
        "SELECT so_test_in1.id, so_test_in1.col1 " \
        "FROM so_test_in1 WHERE so_test_in1.col1 IN " \
        "(SELECT so_test_in2.col2 FROM so_test_in2)"

    select = SOTestIn1.select(IN(SOTestIn1.q.col1, SOTestIn2.select()))
    assert str(select) == \
        "SELECT so_test_in1.id, so_test_in1.col1 " \
        "FROM so_test_in1 WHERE so_test_in1.col1 IN " \
        "(SELECT so_test_in2.id FROM so_test_in2 WHERE 1 = 1)"


def test_2perform_in():
    insert()
    select = SOTestIn1.select(IN(SOTestIn1.q.col1, Select(SOTestIn2.q.col2)))
    assert select.count() == 2


def test_3syntax_exists():
    setup()
    select = SOTestIn1.select(NOTEXISTS(
        Select(SOTestIn2.q.col2,
               where=(Outer(SOTestIn1).q.col1 == SOTestIn2.q.col2))))
    assert str(select) == \
        "SELECT so_test_in1.id, so_test_in1.col1 " \
        "FROM so_test_in1 WHERE NOT EXISTS " \
        "(SELECT so_test_in2.col2 FROM so_test_in2 " \
        "WHERE ((so_test_in1.col1) = (so_test_in2.col2)))"

    setupClass(SOTestOuter)
    select = SOTestOuter.select(NOTEXISTS(
        Select(SOTestIn1.q.col1,
               where=(Outer(SOTestOuter).q.fk == SOTestIn1.q.id))))
    assert str(select) == \
        "SELECT so_test_outer.id, so_test_outer.fk_id " \
        "FROM so_test_outer WHERE NOT EXISTS " \
        "(SELECT so_test_in1.col1 FROM so_test_in1 " \
        "WHERE ((so_test_outer.fk_id) = (so_test_in1.id)))"


def test_4perform_exists():
    insert()
    select = SOTestIn1.select(EXISTS(
        Select(SOTestIn2.q.col2,
               where=(Outer(SOTestIn1).q.col1 == SOTestIn2.q.col2))))
    assert len(list(select)) == 2

    setupClass(SOTestOuter)
    select = SOTestOuter.select(NOTEXISTS(
        Select(SOTestIn1.q.col1,
               where=(Outer(SOTestOuter).q.fkID == SOTestIn1.q.id))))
    assert len(list(select)) == 0


def test_4syntax_direct():
    setup()
    select = SOTestIn1.select(SOTestIn1.q.col1 == Select(SOTestIn2.q.col2,
                              where=(SOTestIn2.q.col2 == "test")))
    assert str(select) == \
        "SELECT so_test_in1.id, so_test_in1.col1 " \
        "FROM so_test_in1 WHERE ((so_test_in1.col1) = " \
        "(SELECT so_test_in2.col2 FROM so_test_in2 " \
        "WHERE ((so_test_in2.col2) = ('test'))))"


def test_4perform_direct():
    insert()
    select = SOTestIn1.select(SOTestIn1.q.col1 == Select(SOTestIn2.q.col2,
                              where=(SOTestIn2.q.col2 == "test")))
    assert select.count() == 1


def test_5perform_direct():
    insert()
    select = SOTestIn1.select(SOTestIn1.q.col1 == Select(SOTestIn2.q.col2,
                              where=(SOTestIn2.q.col2 == "test")))
    assert select.count() == 1


def test_6syntax_join():
    insert()
    j = LEFTOUTERJOINOn(SOTestIn2, SOTestIn1,
                        SOTestIn1.q.col1 == SOTestIn2.q.col2)
    select = SOTestIn1.select(SOTestIn1.q.col1 == Select(SOTestIn2.q.col2,
                              where=(SOTestIn2.q.col2 == "test"), join=j))
    assert str(select) == \
        "SELECT so_test_in1.id, so_test_in1.col1 " \
        "FROM so_test_in1 WHERE ((so_test_in1.col1) = " \
        "(SELECT so_test_in2.col2 FROM so_test_in2 " \
        "LEFT OUTER JOIN so_test_in1 ON " \
        "((so_test_in1.col1) = (so_test_in2.col2)) " \
        "WHERE ((so_test_in2.col2) = ('test'))))"


def test_6perform_join():
    insert()
    j = LEFTOUTERJOINOn(SOTestIn2, SOTestIn1,
                        SOTestIn1.q.col1 == SOTestIn2.q.col2)
    select = SOTestIn1.select(SOTestIn1.q.col1 == Select(SOTestIn2.q.col2,
                              where=(SOTestIn2.q.col2 == "test"), join=j))
    assert select.count() == 1



from sqlobject import SQLObject
from sqlobject.tests.dbtest import setupClass


class SOTestComparison(SQLObject):
    pass


def test_eq():
    setupClass(SOTestComparison, force=True)
    t1 = SOTestComparison()
    t2 = SOTestComparison()

    SOTestComparison._connection.cache.clear()
    t3 = SOTestComparison.get(1)
    t4 = SOTestComparison.get(2)

    assert t1.id == t3.id
    assert t2.id == t4.id
    assert t1 is not t3
    assert t2 is not t4
    assert t1 == t3
    assert t2 == t4
    assert t1 != t2



from decimal import Decimal
import pytest
from sqlobject import DecimalCol, DecimalStringCol, SQLObject, UnicodeCol
from sqlobject.tests.dbtest import setupClass, supports


########################################
# Decimal columns
########################################


try:
    support_decimal_column = supports('decimalColumn')
except NameError:
    # The module was imported during documentation building
    pass
else:
    if not support_decimal_column:
        pytestmark = pytest.mark.skip("These tests require Decimal support")


class DecimalTable(SQLObject):
    name = UnicodeCol(length=255)
    col1 = DecimalCol(size=6, precision=4)
    col2 = DecimalStringCol(size=6, precision=4)
    col3 = DecimalStringCol(size=6, precision=4, quantize=True)


def test_1_decimal():
    setupClass(DecimalTable)
    d = DecimalTable(name='test', col1=21.12, col2='10.01', col3='10.01')
    # psycopg2 returns float as Decimal
    if isinstance(d.col1, Decimal):
        assert d.col1 == Decimal("21.12")
    else:
        assert d.col1 == 21.12
    assert d.col2 == Decimal("10.01")
    assert DecimalTable.sqlmeta.columns['col2'].to_python(
        '10.01', d._SO_validatorState) == Decimal("10.01")
    assert DecimalTable.sqlmeta.columns['col2'].from_python(
        '10.01', d._SO_validatorState) == "10.01"
    assert d.col3 == Decimal("10.01")
    assert DecimalTable.sqlmeta.columns['col3'].to_python(
        '10.01', d._SO_validatorState) == Decimal("10.01")
    assert DecimalTable.sqlmeta.columns['col3'].from_python(
        '10.01', d._SO_validatorState) == "10.0100"


def test_2_decimal():
    setupClass(DecimalTable)
    d = DecimalTable(name='test', col1=Decimal("21.12"),
                     col2=Decimal('10.01'), col3=Decimal('10.01'))
    assert d.col1 == Decimal("21.12")
    assert d.col2 == Decimal("10.01")
    assert DecimalTable.sqlmeta.columns['col2'].to_python(
        Decimal('10.01'), d._SO_validatorState) == Decimal("10.01")
    assert DecimalTable.sqlmeta.columns['col2'].from_python(
        Decimal('10.01'), d._SO_validatorState) == "10.01"
    assert d.col3 == Decimal("10.01")
    assert DecimalTable.sqlmeta.columns['col3'].to_python(
        Decimal('10.01'), d._SO_validatorState) == Decimal("10.01")
    assert DecimalTable.sqlmeta.columns['col3'].from_python(
        Decimal('10.01'), d._SO_validatorState) == "10.0100"



import pytest
from sqlobject import PickleCol, SQLObject
from sqlobject.tests.dbtest import setupClass, supports


########################################
# Pickle columns
########################################


class PickleData:
    pi = 3.14156

    def __init__(self):
        self.question = \
            'The Ulimate Question of Life, the Universe and Everything'
        self.answer = 42


class PickleContainer(SQLObject):
    pickledata = PickleCol(default=None, length=256)


def test_pickleCol():
    if not supports('blobData'):
        pytest.skip("blobData isn't supported")
    setupClass([PickleContainer], force=True)
    mypickledata = PickleData()

    ctnr = PickleContainer(pickledata=mypickledata)
    iid = ctnr.id

    PickleContainer._connection.cache.clear()

    ctnr2 = PickleContainer.get(iid)
    s2 = ctnr2.pickledata

    assert isinstance(s2, PickleData)
    assert isinstance(s2.pi, float)
    assert isinstance(s2.question, str)
    assert isinstance(s2.answer, int)
    assert s2.pi == mypickledata.pi
    assert s2.question == mypickledata.question
    assert s2.answer == mypickledata.answer



from __future__ import print_function
import pytest

from sqlobject import sqlhub, SQLObject, StringCol
try:
    from sqlobject.wsgi_middleware import make_middleware
except ImportError:
    pytestmark = pytest.mark.skipif(True, reason='These tests require Paste')
from .dbtest import getConnection, getConnectionURI, setupClass


class NameOnly(SQLObject):
    name = StringCol()


def makeapp(abort=False, begin=False, fail=False):
    def app(environ, start_response):
        NameOnly(name='app1')
        if fail == 'early':
            assert 0
        start_response('200 OK', [('content-type', 'text/plain')])
        if begin:
            environ['sqlobject.begin']()
        NameOnly(name='app2')
        if abort:
            environ['sqlobject.abort']()
        if fail:
            assert 0
        return ['ok']
    return app


def makestack(abort=False, begin=False, fail=False, **kw):
    app = makeapp(abort=abort, begin=begin, fail=fail)
    app = make_middleware(app, {}, database=getConnectionURI(), **kw)
    return app


def runapp(**kw):
    print('-' * 8)
    app = makestack(**kw)
    env = {}

    def start_response(*args):
        pass

    try:
        list(app(env, start_response))
        return True
    except AssertionError:
        return False


def setup():
    setupClass(NameOnly)
    getConnection().query('DELETE FROM name_only')
    NameOnly._connection = sqlhub


def names():
    names = [n.name for n in NameOnly.select(connection=getConnection())]
    names.sort()
    return names


def test_fail():
    setup()
    assert not runapp(fail=True, use_transaction=True)
    assert names() == []
    setup()
    assert not runapp(fail=True, use_transaction=False)
    assert names() == ['app1', 'app2']
    setup()
    assert not runapp(fail=True, begin=True, use_transaction=True)
    assert names() == ['app1']


def test_other():
    setup()
    assert runapp(fail=False, begin=True, use_transaction=True)
    assert names() == ['app1', 'app2']
    setup()
    # @@: Dammit, I can't get these to pass because I can't get the
    # stupid table to clear itself.  setupClass() sucks.  When I
    # fix it I'll take this disabling out:
    pytest.skip("Oops...")
    assert names() == []
    assert runapp(fail=False, begin=True, abort=True, use_transaction=True)
    assert names() == ['app1']
    setup()
    assert runapp(use_transaction=True)
    assert names() == ['app1', 'app2']



from __future__ import print_function
import sys
import types
from pydispatch import dispatcher
from weakref import ref
from .compat import class_types


subclassClones = {}


def listen(receiver, soClass, signal, alsoSubclasses=True, weak=True):
    """
    Listen for the given ``signal`` on the SQLObject subclass
    ``soClass``, calling ``receiver()`` when ``send(soClass, signal,
    ...)`` is called.

    If ``alsoSubclasses`` is true, receiver will also be called when
    an event is fired on any subclass.
    """
    dispatcher.connect(receiver, signal=signal, sender=soClass, weak=weak)
    weakReceiver = ref(receiver)
    subclassClones.setdefault(soClass, []).append((weakReceiver, signal))

# We export this function:
send = dispatcher.send


class Signal(object):
    """
    Base event for all SQLObject events.

    In general the sender for these methods is the class, not the
    instance.
    """


class ClassCreateSignal(Signal):
    """
    Signal raised after class creation.  The sender is the superclass
    (in case of multiple superclasses, the first superclass).  The
    arguments are ``(new_class_name, bases, new_attrs, post_funcs,
    early_funcs)``.  ``new_attrs`` is a dictionary and may be modified
    (but ``new_class_name`` and ``bases`` are immutable).
    ``post_funcs`` is an initially-empty list that can have callbacks
    appended to it.

    Note: at the time this event is called, the new class has not yet
    been created.  The functions in ``post_funcs`` will be called
    after the class is created, with the single arguments of
    ``(new_class)``.  Also, ``early_funcs`` will be called at the
    soonest possible time after class creation (``post_funcs`` is
    called after the class's ``__classinit__``).
    """


def _makeSubclassConnections(new_class_name, bases, new_attrs,
                             post_funcs, early_funcs):
    early_funcs.insert(0, _makeSubclassConnectionsPost)


def _makeSubclassConnectionsPost(new_class):
    for cls in new_class.__bases__:
        for weakReceiver, signal in subclassClones.get(cls, []):
            receiver = weakReceiver()
            if not receiver:
                continue
            listen(receiver, new_class, signal)

dispatcher.connect(_makeSubclassConnections, signal=ClassCreateSignal)


# @@: Should there be a class reload event?  This would allow modules
# to be reloaded, possibly.  Or it could even be folded into
# ClassCreateSignal, since anything that listens to that needs to pay
# attention to reloads (or else it is probably buggy).


class RowCreateSignal(Signal):
    """
    Called before an instance is created, with the class as the
    sender.  Called with the arguments ``(instance, kwargs, post_funcs)``.
    There may be a ``connection`` argument.  ``kwargs``may be usefully
    modified.  ``post_funcs`` is a list of callbacks, intended to have
    functions appended to it, and are called with the arguments
    ``(new_instance)``.

    Note: this is not called when an instance is created from an
    existing database row.
    """


class RowCreatedSignal(Signal):
    """
    Called after an instance is created, with the class as the
    sender.  Called with the arguments ``(instance, kwargs, post_funcs)``.
    There may be a ``connection`` argument.  ``kwargs``may be usefully
    modified.  ``post_funcs`` is a list of callbacks, intended to have
    functions appended to it, and are called with the arguments
    ``(new_instance)``.

    Note: this is not called when an instance is created from an
    existing database row.
    """
# @@: An event for getting a row?  But for each row, when doing a
# select?  For .sync, .syncUpdate, .expire?


class RowDestroySignal(Signal):
    """
    Called before an instance is deleted.  Sender is the instance's
    class.  Arguments are ``(instance, post_funcs)``.

    ``post_funcs`` is a list of callbacks, intended to have
    functions appended to it, and are called with arguments ``(instance)``.
    If any of the post_funcs raises an exception, the deletion is only
    affected if this will prevent a commit.

    You cannot cancel the delete, but you can raise an exception (which will
    probably cancel the delete, but also cause an uncaught exception if not
    expected).

    Note: this is not called when an instance is destroyed through
    garbage collection.

    @@: Should this allow ``instance`` to be a primary key, so that a
    row can be deleted without first fetching it?
    """


class RowDestroyedSignal(Signal):
    """
    Called after an instance is deleted.  Sender is the instance's
    class.  Arguments are ``(instance)``.

    This is called before the post_funcs of RowDestroySignal

    Note: this is not called when an instance is destroyed through
    garbage collection.
    """


class RowUpdateSignal(Signal):
    """
    Called when an instance is updated through a call to ``.set()``
    (or a column attribute assignment).  The arguments are
    ``(instance, kwargs)``.  ``kwargs`` can be modified.  This is run
    *before* the instance is updated; if you want to look at the
    current values, simply look at ``instance``.
    """


class RowUpdatedSignal(Signal):
    """
    Called when an instance is updated through a call to ``.set()``
    (or a column attribute assignment).  The arguments are
    ``(instance, post_funcs)``. ``post_funcs`` is a list of callbacks,
    intended to have functions appended to it, and are called with the
    arguments ``(new_instance)``. This is run *after* the instance is
    updated; Works better with lazyUpdate = True.
    """


class AddColumnSignal(Signal):
    """
    Called when a column is added to a class, with arguments ``(cls,
    connection, column_name, column_definition, changeSchema,
    post_funcs)``.  This is called *after* the column has been added,
    and is called for each column after class creation.

    post_funcs are called with ``(cls, so_column_obj)``
    """


class DeleteColumnSignal(Signal):
    """
    Called when a column is removed from a class, with the arguments
    ``(cls, connection, column_name, so_column_obj, post_funcs)``.
    Like ``AddColumnSignal`` this is called after the action has been
    performed, and is called for subclassing (when a column is
    implicitly removed by setting it to ``None``).

    post_funcs are called with ``(cls, so_column_obj)``
    """

# @@: Signals for indexes and joins?  These are mostly event consumers,
# though.


class CreateTableSignal(Signal):
    """
    Called when a table is created.  If ``ifNotExists==True`` and the
    table exists, this event is not called.

    Called with ``(cls, connection, extra_sql, post_funcs)``.
    ``extra_sql`` is a list (which can be appended to) of extra SQL
    statements to be run after the table is created.  ``post_funcs``
    functions are called with ``(cls, connection)`` after the table
    has been created.  Those functions are *not* called simply when
    constructing the SQL.
    """


class DropTableSignal(Signal):
    """
    Called when a table is dropped.  If ``ifExists==True`` and the
    table doesn't exist, this event is not called.

    Called with ``(cls, connection, extra_sql, post_funcs)``.
    ``post_funcs`` functions are called with ``(cls, connection)``
    after the table has been dropped.
    """


class CommitSignal(Signal):
    """
    Called on transaction commit
    """


class RollbackSignal(Signal):
    """
    Called on transaction rollback
    """


############################################################
# Event Debugging
############################################################


def summarize_events_by_sender(sender=None, output=None, indent=0):
    """
    Prints out a summary of the senders and listeners in the system,
    for debugging purposes.
    """
    if output is None:
        output = sys.stdout
    leader = ' ' * indent
    if sender is None:
        send_list = [
            (deref(dispatcher.senders.get(sid)), listeners)
            for sid, listeners in dispatcher.connections.items()
            if deref(dispatcher.senders.get(sid))]
        for sender, listeners in sorted_items(send_list):
            real_sender = deref(sender)
            if not real_sender:
                continue
            header = 'Sender: %r' % real_sender
            print(leader + header, file=output)
            print(leader + ('=' * len(header)), file=output)
            summarize_events_by_sender(real_sender, output=output,
                                       indent=indent + 2)
    else:
        for signal, receivers in \
                sorted_items(dispatcher.connections.get(id(sender), [])):
            receivers = [deref(r) for r in receivers if deref(r)]
            header = 'Signal: %s (%i receivers)' % (sort_name(signal),
                                                    len(receivers))
            print(leader + header, file=output)
            print(leader + ('-' * len(header)), file=output)
            for receiver in sorted(receivers, key=sort_name):
                print(leader + '  ' + nice_repr(receiver), file=output)


def deref(value):
    if isinstance(value, dispatcher.WEAKREF_TYPES):
        return value()
    else:
        return value


def sorted_items(a_dict):
    if isinstance(a_dict, dict):
        a_dict = a_dict.items()
    return sorted(a_dict, key=lambda t: sort_name(t[0]))


def sort_name(value):
    if isinstance(value, type):
        return value.__name__
    elif isinstance(value, types.FunctionType):
        return value.__name__
    else:
        return str(value)

_real_dispatcher_send = dispatcher.send
_real_dispatcher_sendExact = dispatcher.sendExact
_real_dispatcher_connect = dispatcher.connect
_real_dispatcher_disconnect = dispatcher.disconnect
_debug_enabled = False


def debug_events():
    global _debug_enabled, send
    if _debug_enabled:
        return
    _debug_enabled = True
    dispatcher.send = send = _debug_send
    dispatcher.sendExact = _debug_sendExact
    dispatcher.disconnect = _debug_disconnect
    dispatcher.connect = _debug_connect


def _debug_send(signal=dispatcher.Any, sender=dispatcher.Anonymous,
                *arguments, **named):
    print("send %s from %s: %s" % (
          nice_repr(signal), nice_repr(sender),
          fmt_args(*arguments, **named)))
    return _real_dispatcher_send(signal, sender, *arguments, **named)


def _debug_sendExact(signal=dispatcher.Any, sender=dispatcher.Anonymous,
                     *arguments, **named):
    print("sendExact %s from %s: %s" % (
          nice_repr(signal), nice_repr(sender), fmt_args(*arguments, **name)))
    return _real_dispatcher_sendExact(signal, sender, *arguments, **named)


def _debug_connect(receiver, signal=dispatcher.Any, sender=dispatcher.Any,
                   weak=True):
    print("connect %s to %s signal %s" % (
          nice_repr(receiver), nice_repr(signal), nice_repr(sender)))
    return _real_dispatcher_connect(receiver, signal, sender, weak)


def _debug_disconnect(receiver, signal=dispatcher.Any, sender=dispatcher.Any,
                      weak=True):
    print("disconnecting %s from %s signal %s" % (
          nice_repr(receiver), nice_repr(signal), nice_repr(sender)))
    return _real_dispatcher_disconnect(receiver, signal, sender, weak)


def fmt_args(*arguments, **name):
    args = [repr(a) for a in arguments]
    args.extend([
        '%s=%r' % (n, v) for n, v in sorted(name.items())])
    return ', '.join(args)


def nice_repr(v):
    """
    Like repr(), but nicer for debugging here.
    """
    if isinstance(v, class_types):
        return v.__module__ + '.' + v.__name__
    elif isinstance(v, types.FunctionType):
        if '__name__' in v.__globals__:
            if getattr(sys.modules[v.__globals__['__name__']],
                       v.__name__, None) is v:
                return '%s.%s' % (v.__globals__['__name__'], v.__name__)
        return repr(v)
    elif isinstance(v, types.MethodType):
        return '%s.%s of %s' % (
            nice_repr(v.__self__.__class__), v.__func__.__name__,
            nice_repr(v.__self__))
    else:
        return repr(v)


__all__ = ['listen', 'send']
# Use copy() to avoid 'dictionary changed' issues on python 3
for name, value in globals().copy().items():
    if isinstance(value, type) and issubclass(value, Signal):
        __all__.append(name)



from sqlobject import EnumCol, SQLObject, UnicodeCol
from sqlobject.col import validators
from sqlobject.tests.dbtest import raises, setupClass


########################################
# Enum test
########################################


class Enum1(SQLObject):

    cl = EnumCol(enumValues=['a', 'bcd', 'e'])


def testBad():
    setupClass(Enum1)
    for _l in ['a', 'bcd', 'a', 'e']:
        Enum1(cl=_l)
    raises(
        (Enum1._connection.module.IntegrityError,
         Enum1._connection.module.ProgrammingError,
         validators.Invalid),
        Enum1, cl='b')


class EnumWithNone(SQLObject):

    cl = EnumCol(enumValues=['a', 'bcd', 'e', None])


def testNone():
    setupClass(EnumWithNone)
    for _l in [None, 'a', 'bcd', 'a', 'e', None]:
        e = EnumWithNone(cl=_l)
        assert e.cl == _l


class EnumWithDefaultNone(SQLObject):

    cl = EnumCol(enumValues=['a', 'bcd', 'e', None], default=None)


def testDefaultNone():
    setupClass(EnumWithDefaultNone)

    e = EnumWithDefaultNone()
    assert e.cl is None


class EnumWithDefaultOther(SQLObject):

    cl = EnumCol(enumValues=['a', 'bcd', 'e', None], default='a')


def testDefaultOther():
    setupClass(EnumWithDefaultOther)

    e = EnumWithDefaultOther()
    assert e.cl == 'a'


class EnumUnicode(SQLObject):

    n = UnicodeCol()
    cl = EnumCol(enumValues=['a', 'b'])


def testUnicode():
    setupClass(EnumUnicode)

    EnumUnicode(n=u'a', cl='a')
    EnumUnicode(n=u'b', cl=u'b')
    EnumUnicode(n=u'\u201c', cl='a')
    EnumUnicode(n=u'\u201c', cl=u'b')



"""
classresolver.py
  2 February 2004, Ian Bicking <ianb@colorstudy.com>

Resolves strings to classes, and runs callbacks when referenced
classes are created.

Classes are referred to only by name, not by module.  So that
identically-named classes can coexist, classes are put into individual
registries, which are keyed on strings (names).  These registries are
created on demand.

Use like::

    >>> import classregistry
    >>> registry = classregistry.registry('MyModules')
    >>> def afterMyClassExists(cls):
    ...    print('Class finally exists: %s' % cls)
    >>> registry.addClassCallback('MyClass', afterMyClassExists)
    >>> class MyClass:
    ...    pass
    >>> registry.addClass(MyClass)
    Class finally exists: MyClass

"""


class ClassRegistry(object):
    """
    We'll be dealing with classes that reference each other, so
    class C1 may reference C2 (in a join), while C2 references
    C1 right back.  Since classes are created in an order, there
    will be a point when C1 exists but C2 doesn't.  So we deal
    with classes by name, and after each class is created we
    try to fix up any references by replacing the names with
    actual classes.

    Here we keep a dictionaries of class names to classes -- note
    that the classes might be spread among different modules, so
    since we pile them together names need to be globally unique,
    to just module unique.
    Like needSet below, the container dictionary is keyed by the
    class registry.
    """

    def __init__(self, name):
        self.name = name
        self.classes = {}
        self.callbacks = {}
        self.genericCallbacks = []

    def addClassCallback(self, className, callback, *args, **kw):
        """
        Whenever a name is substituted for the class, you can register
        a callback that will be called when the needed class is
        created.  If it's already been created, the callback will be
        called immediately.
        """
        if className in self.classes:
            callback(self.classes[className], *args, **kw)
        else:
            self.callbacks.setdefault(className, []).append(
                (callback, args, kw))

    def addCallback(self, callback, *args, **kw):
        """
        This callback is called for all classes, not just specific
        ones (like addClassCallback).
        """
        self.genericCallbacks.append((callback, args, kw))
        for cls in self.classes.values():
            callback(cls, *args, **kw)

    def addClass(self, cls):
        """
        Everytime a class is created, we add it to the registry, so
        that other classes can find it by name.  We also call any
        callbacks that are waiting for the class.
        """
        if cls.__name__ in self.classes:
            import sys
            other = self.classes[cls.__name__]
            raise ValueError(
                "class %s is already in the registry (other class is "
                "%r, from the module %s in %s; attempted new class is "
                "%r, from the module %s in %s)"
                % (cls.__name__,
                   other, other.__module__,
                   getattr(sys.modules.get(other.__module__),
                           '__file__', '(unknown)'),
                   cls, cls.__module__,
                   getattr(sys.modules.get(cls.__module__),
                           '__file__', '(unknown)')))
        self.classes[cls.__name__] = cls
        if cls.__name__ in self.callbacks:
            for callback, args, kw in self.callbacks[cls.__name__]:
                callback(cls, *args, **kw)
            del self.callbacks[cls.__name__]
        for callback, args, kw in self.genericCallbacks:
            callback(cls, *args, **kw)

    def getClass(self, className):
        try:
            return self.classes[className]
        except KeyError:
            all = sorted(self.classes.keys())
            raise KeyError(
                "No class %s found in the registry %s (these classes "
                "exist: %s)"
                % (className, self.name or '[default]', ', '.join(all)))

    def allClasses(self):
        return self.classes.values()


class _MasterRegistry(object):
    """
    This singleton holds all the class registries.  There can be
    multiple registries to hold different unrelated sets of classes
    that reside in the same process.  These registries are named with
    strings, and are created on demand.  The MasterRegistry module
    global holds the singleton.
    """

    def __init__(self):
        self.registries = {}

    def registry(self, item):
        if item not in self.registries:
            self.registries[item] = ClassRegistry(item)
        return self.registries[item]

MasterRegistry = _MasterRegistry()
registry = MasterRegistry.registry


def findClass(name, class_registry=None):
    return registry(class_registry).getClass(name)



from sqlobject import ForeignKey, SQLObject, StringCol, \
    SQLObjectIntegrityError, SQLObjectNotFound
from sqlobject.tests.dbtest import raises, setupClass


class SOTestPerson1(SQLObject):
    name = StringCol()


class SOTestMessageCascadeTrue(SQLObject):
    sender = ForeignKey('SOTestPerson1', cascade=True)
    recipient = ForeignKey('SOTestPerson1', cascade=True)
    body = StringCol()


def test1():
    setupClass([SOTestPerson1, SOTestMessageCascadeTrue])

    john = SOTestPerson1(name='john')
    emily = SOTestPerson1(name='emily')
    message = SOTestMessageCascadeTrue(
        sender=emily, recipient=john, body='test1'
    )

    SOTestPerson1.delete(emily.id)
    john.expire()
    message.expire()

    john.sync()
    raises(SQLObjectNotFound, emily.sync)
    raises(SQLObjectNotFound, message.sync)


class SOTestPerson2(SQLObject):
    name = StringCol()


class SOTestMessageCascadeFalse(SQLObject):
    sender = ForeignKey('SOTestPerson2', cascade=False)
    recipient = ForeignKey('SOTestPerson2', cascade=False)
    body = StringCol()


def test2():
    setupClass([SOTestPerson2, SOTestMessageCascadeFalse])

    john = SOTestPerson2(name='john')
    emily = SOTestPerson2(name='emily')
    message = SOTestMessageCascadeFalse(
        sender=emily, recipient=john, body='test2'
    )

    raises(SQLObjectIntegrityError, SOTestPerson2.delete, emily.id)
    john.expire()
    emily.expire()
    message.expire()

    john.sync()
    emily.sync()
    message.sync()

    assert message.sender == emily
    assert message.recipient == john


class SOTestPerson3(SQLObject):
    name = StringCol()


class SOTestMessageCascadeNull(SQLObject):
    sender = ForeignKey('SOTestPerson3', cascade='null')
    recipient = ForeignKey('SOTestPerson3', cascade='null')
    body = StringCol()


def test3():
    setupClass([SOTestPerson3, SOTestMessageCascadeNull])

    john = SOTestPerson3(name='john')
    emily = SOTestPerson3(name='emily')
    message = SOTestMessageCascadeNull(
        sender=emily, recipient=john, body='test3'
    )

    SOTestPerson3.delete(emily.id)
    john.expire()
    message.expire()

    john.sync()
    message.sync()
    raises(SQLObjectNotFound, emily.sync)

    assert message.sender is None
    assert message.recipient == john

    SOTestPerson3.delete(john.id)
    john.expire()
    message.expire()

    message.sync()
    raises(SQLObjectNotFound, john.sync)

    assert message.recipient is None


class SOTestPerson4(SQLObject):
    name = StringCol()


class SOTestMessageCascadeMixed(SQLObject):
    sender = ForeignKey('SOTestPerson4', cascade=True)
    recipient = ForeignKey('SOTestPerson4', cascade='null')
    body = StringCol()


def test4():
    setupClass([SOTestPerson4, SOTestMessageCascadeMixed])

    john = SOTestPerson4(name='john')
    emily = SOTestPerson4(name='emily')
    message = SOTestMessageCascadeMixed(
        sender=emily, recipient=john, body='test4'
    )

    SOTestPerson4.delete(emily.id)
    john.expire()
    message.expire()

    john.sync()
    raises(SQLObjectNotFound, message.sync)


def test5():
    setupClass([SOTestPerson4, SOTestMessageCascadeMixed])

    john = SOTestPerson4(name='john')
    emily = SOTestPerson4(name='emily')
    message = SOTestMessageCascadeMixed(
        sender=emily, recipient=john, body='test5'
    )

    john.destroySelf()
    emily.expire()
    message.expire()

    emily.sync()
    message.sync()

    assert message.recipient is None
    assert message.sender == emily



from sqlobject.compat import PY2
from sqlobject.constraints import BadValue, InList, MaxLength, \
    isFloat, isInt, isString, notNull
from sqlobject.tests.dbtest import Dummy, raises

if not PY2:
    # alias for python 3 compatability
    long = int


def test_constraints():
    obj = 'Test object'
    col = Dummy(name='col')
    isString(obj, col, 'blah')
    raises(BadValue, isString, obj, col, 1)
    if PY2:
        # @@: Should this really be an error?
        raises(BadValue, isString, obj, col, u'test!')
    else:
        raises(BadValue, isString, obj, col, b'test!')
    # isString(obj, col, u'test!')

    raises(BadValue, notNull, obj, col, None)
    raises(BadValue, isInt, obj, col, 1.1)
    isInt(obj, col, 1)
    isInt(obj, col, long(1))
    isFloat(obj, col, 1)
    isFloat(obj, col, long(1))
    isFloat(obj, col, 1.2)
    raises(BadValue, isFloat, obj, col, '1.0')

    # @@: Should test isBool, but I don't think isBool is right

    lst = InList(('a', 'b', 'c'))
    lst(obj, col, 'a')
    raises(BadValue, lst, obj, col, ('a', 'b', 'c'))
    raises(BadValue, lst, obj, col, 'A')

    maxlen = MaxLength(2)
    raises(BadValue, maxlen, obj, col, '123')
    maxlen(obj, col, '12')
    maxlen(obj, col, (1,))
    raises(BadValue, maxlen, obj, col, 1)



from sqlobject import SQLObject, StringCol
from sqlobject.tests.dbtest import setupClass


########################################
# sqlmeta.asDict()
########################################


class SOTestAsDict(SQLObject):
    name = StringCol(length=10)
    name2 = StringCol(length=10)


def test_asDict():
    setupClass(SOTestAsDict, force=True)
    t1 = SOTestAsDict(name='one', name2='1')
    assert t1.sqlmeta.asDict() == dict(name='one', name2='1', id=1)
